import unittest
from unittest.mock import MagicMock

from palaestrai.experiment import RunGovernor
from palaestrai.experiment.rungovernor_states import (
    RGSErrorHandlingStarting,
    RGSStoppingSimulation,
    RGSStoppingTransceiving,
    RGSTransceiving,
)
from palaestrai.util.exception import (
    AgentConductorFailedError,
    EnvConductorFailedError,
    ExperimentAlreadyRunningError,
    ExperimentSetupFailedError,
    InvalidResponseError,
    SimControllerFailedError,
)


class TestHandlingErrorsStarting(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.rungov = RunGovernor(None, None)
        self.rgs = RGSErrorHandlingStarting(self.rungov)
        self.rungov.state = self.rgs

    async def test_run(self):
        logger = f"{RunGovernor.__module__}".rsplit(".", 1)[0]

        self.rgs.add_error(ExperimentAlreadyRunningError())

        with self.assertLogs(logger, level="DEBUG") as cm:
            await self.rgs.run()

        self.assertIn("Handling these errors now.", cm.output[0])

    def test_next_state_running_experiment(self):

        self.rgs.add_error(ExperimentAlreadyRunningError())
        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSTransceiving)

    def test_next_state_failed_setup_experiment(self):

        self.rgs.add_error(ExperimentSetupFailedError())
        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSStoppingTransceiving)

    def test_next_state_simcontroller_failed(self):
        self.rgs.add_error(SimControllerFailedError())
        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSStoppingSimulation)

    def test_next_state_env_conductor_failed(self):
        self.rgs.add_error(EnvConductorFailedError())
        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSStoppingSimulation)

    def test_next_state_invalid_response_error(self):
        self.rgs.add_error(
            InvalidResponseError(expected=MagicMock(), got=MagicMock())
        )
        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSStoppingSimulation)

    def test_next_state_agent_conductor_failed(self):
        self.rgs.add_error(AgentConductorFailedError())
        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSStoppingSimulation)



from __future__ import annotations
from dataclasses import dataclass
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from palaestrai.agent import ActuatorInformation


@dataclass
class EnvironmentUpdateRequest:
    """Updates an environment with new set points from actuators.

    * Sender: :class:`SimulationController`
    * Receiver: :class:`Environment`

    :param sender_simulation_controller: ID of the sending
        :class:`SimulationController`
    :param receiver_environment: ID of the receiving environment
    :param experiment_run_id: ID of the current experiment run this environment
        participates in
    :param environment_conductor_id: ID of the :class:`EnvironmentConductor`
        the environment belongs to
    :param actuators: list of :class:`ActuatorInformation` objects that convey
        new setpoints to the environment
    """

    sender_simulation_controller: str
    receiver_environment: str
    experiment_run_id: str
    environment_conductor_id: str
    actuators: List[ActuatorInformation]

    @property
    def sender(self):
        return self.sender_simulation_controller

    @property
    def receiver(self):
        return self.receiver_environment

    @property
    def environment_id(self):
        return self.receiver_environment



import unittest

from palaestrai.experiment import RunGovernor
from palaestrai.experiment.rungovernor_states import RunGovernorState


class DummyState(RunGovernorState):
    def __init__(self, rgc, called=0):
        super().__init__(rgc, "TEST")
        self.called = called

    async def run(self):
        self.called += 1

        if self.called >= 3:
            self.rgc.shutdown = True

    def next_state(self):
        if self.called == 2:
            self.rgc.state = DummyState(self.rgc, self.called)


class TestRunGovernor(unittest.IsolatedAsyncioTestCase):
    async def test_run_state_change(self):
        """Test the run method.
        A test state is created, which transitions to another test
        state after being called two times.

        """
        rungov = RunGovernor("test_broker_uri", None)
        test_state = DummyState(rungov)
        rungov.state = test_state

        await rungov.run()

        self.assertEqual(2, test_state.called)
        self.assertEqual(3, rungov.state.called)


if __name__ == "__main__":
    unittest.main()



import sqlalchemy as sa
from sqlalchemy import func
from sqlalchemy.orm import relationship
from palaestrai.store.database_base import Base

"""The database schema provided as Model for SQLAlchemy"""


class ExperimentStudy(Base):
    __tablename__ = "experiment_studies"
    id = sa.Column(sa.INTEGER, primary_key=True, unique=True, index=True)
    name = sa.Column(sa.String, nullable=True)
    study = sa.Column(sa.TEXT)
    experiments = relationship(
        "Experiment", back_populates="experiment_relation"
    )


class Experiment(Base):
    __tablename__ = "experiments"
    id = sa.Column(sa.Integer, primary_key=True, unique=True, index=True)
    name = sa.Column(sa.VARCHAR(255), unique=True, index=True)
    experiment_study_id = sa.Column(
        sa.Integer, sa.ForeignKey(ExperimentStudy.id), index=True
    )
    experiment = sa.Column(sa.TEXT)
    experiment_relation = relationship(
        "ExperimentStudy", back_populates="experiments"
    )
    experiment_runs = relationship(
        "ExperimentRun", back_populates="experiment_run_relation"
    )


class ExperimentRun(Base):
    __tablename__ = "experiment_runs"
    id = sa.Column(sa.Integer, primary_key=True, unique=True, index=True)
    uuid = sa.Column(sa.VARCHAR(255), unique=True, index=True)
    experiment_id = sa.Column(
        sa.Integer, sa.ForeignKey(Experiment.id), index=True
    )
    experiment_run_relation = relationship(
        "Experiment", back_populates="experiment_runs"
    )
    agent_conductors = relationship(
        "AgentConductor", back_populates="agent_conductor_relation"
    )
    simulation_instances = relationship(
        "SimulationInstance", back_populates="simulation_instance_relation"
    )


class SimulationInstance(Base):
    __tablename__ = "simulation_instances"
    id = sa.Column(sa.INTEGER, primary_key=True, unique=True, index=True)
    uuid = sa.Column(sa.VARCHAR(255), unique=True, index=True)
    experiment_run_id = sa.Column(
        sa.Integer, sa.ForeignKey(ExperimentRun.id), index=True
    )
    simulation_instance_relation = relationship(
        "ExperimentRun", back_populates="simulation_instances"
    )
    environment_conductors = relationship(
        "EnvironmentConductor", back_populates="simulation_instance"
    )
    muscles = relationship("Muscle", back_populates="simulation_instance")


class EnvironmentConductor(Base):
    __tablename__ = "environment_conductors"
    id = sa.Column(sa.Integer, primary_key=True, unique=True, index=True)
    uuid = sa.Column(sa.VARCHAR(255), nullable=False, unique=True, index=True)
    type = sa.Column(sa.VARCHAR(255), nullable=True)
    parameters = sa.Column(sa.Text, nullable=True)
    simulation_instance_id = sa.Column(
        sa.Integer,
        sa.ForeignKey(SimulationInstance.id),
        index=True,
    )
    simulation_instance = relationship(
        "SimulationInstance", back_populates="environment_conductors"
    )
    world_states = relationship(
        "WorldState", back_populates="environment_conductor"
    )


class WorldState(Base):
    __tablename__ = "world_states"
    id = sa.Column(sa.Integer, primary_key=True, unique=True, index=True)
    simtime = sa.Column(sa.Integer)
    state_dump = sa.Column(sa.Text)
    is_terminal = sa.Column(
        sa.Boolean,
        unique=False,
        nullable=False,
        default=bool(False),
    )
    environment_conductor_id = sa.Column(
        sa.Integer, sa.ForeignKey(EnvironmentConductor.id), index=True
    )
    environment_conductor = relationship(
        "EnvironmentConductor", back_populates="world_states"
    )


class AgentConductor(Base):
    __tablename__ = "agent_conductors"
    id = sa.Column(sa.Integer, primary_key=True, unique=True, index=True)
    uuid = sa.Column(sa.VARCHAR(255), nullable=False, unique=True, index=True)
    name = sa.Column(sa.VARCHAR(255), nullable=True)
    configuration = sa.Column(sa.LargeBinary, nullable=True)
    experiment_run_id = sa.Column(sa.Integer, sa.ForeignKey(ExperimentRun.id))
    agent_conductor_relation = relationship(
        "ExperimentRun", back_populates="agent_conductors"
    )
    muscles = relationship("Muscle", back_populates="agent_conductor")
    brain = relationship("Brain", back_populates="agent_conductor")


class Brain(Base):
    __tablename__ = "brains"
    id = sa.Column(sa.Integer, primary_key=True, unique=True, index=True)
    parameters = sa.Column(sa.LargeBinary, nullable=True)
    agent_conductor_id = sa.Column(
        sa.Integer, sa.ForeignKey(AgentConductor.id), index=True
    )
    agent_conductor = relationship("AgentConductor", back_populates="brain")
    brain_configs = relationship(
        "BrainConfiguration", back_populates="brain_config_relation"
    )


class BrainConfiguration(Base):
    __tablename__ = "brain_configurations"
    id = sa.Column(sa.Integer, primary_key=True, unique=True, index=True)
    brain_id = sa.Column(sa.Integer, sa.ForeignKey(Brain.id), index=True)
    agent_id = sa.Column(sa.Integer, index=True)
    simtime = sa.Column(sa.Integer)
    parameters = sa.Column(sa.LargeBinary, nullable=True)
    brain_config_relation = relationship(
        "Brain", back_populates="brain_configs"
    )


class Muscle(Base):
    __tablename__ = "muscles"
    id = sa.Column(sa.Integer, primary_key=True, unique=True, index=True)
    agent_conductor_id = sa.Column(
        sa.Integer, sa.ForeignKey(AgentConductor.id), index=True
    )
    simulation_instance_id = sa.Column(
        sa.Integer, sa.ForeignKey(SimulationInstance.id), index=True
    )
    uuid = sa.Column(sa.VARCHAR(255), nullable=False, unique=True, index=True)
    agent_conductor = relationship("AgentConductor", back_populates="muscles")
    simulation_instance = relationship(
        "SimulationInstance", back_populates="muscles"
    )
    muscle_states = relationship(
        "MuscleState", back_populates="muscle_state_relation"
    )
    muscle_actions = relationship("MuscleAction", back_populates="muscle")
    muscle_sensor_readings = relationship(
        "MuscleSensorReading", back_populates="muscle_sensor_reads_relation"
    )


class MuscleState(Base):
    __tablename__ = "muscle_states"
    id = sa.Column(sa.Integer, primary_key=True, unique=True, index=True)
    muscle_id = sa.Column(sa.Integer, sa.ForeignKey(Muscle.id), index=True)
    parameters = sa.Column(sa.LargeBinary, nullable=True)
    muscle_state_relation = relationship(
        "Muscle", back_populates="muscle_states"
    )


class MuscleAction(Base):
    __tablename__ = "muscle_actions"
    id = sa.Column(sa.Integer, primary_key=True, unique=True, index=True)
    simtime = sa.Column(sa.BIGINT)  # TODO: REPLACE, JUST A WORKAROUND
    sensor_readings = sa.Column(sa.Text)
    actuator_information = sa.Column(sa.Text)
    reward = sa.Column(sa.FLOAT)
    muscle_id = sa.Column(sa.Integer, sa.ForeignKey(Muscle.id), index=True)
    muscle = relationship("Muscle", back_populates="muscle_actions")


class MuscleSensorReading(Base):
    __tablename__ = "muscle_sensor_readings"
    id = sa.Column(sa.Integer, primary_key=True, unique=True, index=True)
    active_agent_id = sa.Column(sa.Integer)
    sensor_id = sa.Column(sa.Integer)
    simtime = sa.Column(sa.Integer)
    value = sa.Column(sa.LargeBinary)
    muscle_id = sa.Column(sa.Integer, sa.ForeignKey(Muscle.id), index=True)
    muscle_sensor_reads_relation = relationship(
        "Muscle", back_populates="muscle_sensor_readings"
    )


Model = Base



import pickle
import zlib


def serialize(request):
    pick = pickle.dumps(request)
    return zlib.compress(pick)


def deserialize(response):
    if response is not None:
        decompress = zlib.decompress(response[0])
        return pickle.loads(decompress)
    else:
        return None



from __future__ import annotations

from os import getcwd
from typing import Union, TextIO

import appdirs
import ruamel.yaml

from . import LOG


class _RuntimeConfig:
    """Application-wide runtime configuration.

    This singleton class provides an application-wide runtime configuration
    and transparently hides all sources from the rest of the application.
    """

    CONFIG_FILE_PATHS = [
        "%s/runtime-conf.yaml"
        % appdirs.site_config_dir("palaestrai", "OFFIS"),
        "%s/runtime-conf.yaml"
        % appdirs.user_config_dir("palaestrai", "OFFIS"),
        "%s/palaestrai-runtime.conf.yaml" % getcwd(),
        "%s/runtime.conf.yaml" % getcwd(),
    ]
    DEFAULT_CONFIG = {
        "store_uri": "sqlite:///palaestrai.db",
        "executor_bus_port": 4242,
        "logger_port": 4243,
        "public_bind": False,
        "major_domo_client_timeout": 300_000,
        "major_domo_client_retries": 3,
        "logging": {
            "version": 1,
            "formatters": {
                "simple": {
                    "format": "%(asctime)s %(name)s[%(process)d]: "
                    "%(levelname)s - %(message)s"
                },
                "debug": {
                    "format": "%(asctime)s %(name)s[%(process)d]: "
                    "%(levelname)s - %(message)s (%(module)s.%(funcName)s "
                    "in %(filename)s:%(lineno)d)"
                },
            },
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "level": "INFO",
                    "formatter": "simple",
                    "stream": "ext://sys.stdout",
                },
                "console_debug": {
                    "class": "logging.StreamHandler",
                    "level": "DEBUG",
                    "formatter": "debug",
                    "stream": "ext://sys.stdout",
                },
            },
            "loggers": {
                "palaestrai.agent": {"level": "ERROR"},
                "palaestrai.agent.brain": {"level": "ERROR"},
                "palaestrai.agent.muscle": {"level": "ERROR"},
                "palaestrai.core": {"level": "ERROR"},
                "palaestrai.experiment": {"level": "ERROR"},
                "palaestrai.store": {"level": "ERROR"},
                "palaestrai.environment": {"level": "ERROR"},
                "palaestrai.simulation": {"level": "ERROR"},
                "palaestrai.types": {"level": "ERROR"},
                "palaestrai.util": {"level": "ERROR"},
                "palaestrai.visualization": {"level": "ERROR"},
            },
            "root": {
                "level": "ERROR",
                "handlers": ["console", "console_debug"],
            },
        },
        "profile": False,
    }
    _instance = None

    def __init__(self):
        self._config_file_path = None
        self.config_search_path = _RuntimeConfig.CONFIG_FILE_PATHS

        # The loaded configuration is what RuntimeConfig.load gave us. It
        # remains immutable after loading.
        self._loaded_configuration = {}

    def _get(self, key, default=None, exception=None):
        """Retrieves a config key

        Retrieves any config key; if not set, it queries the config dictionary;
        if it isn't present there, it returns the given default value. It also
        sets the value in the current object as a side-effect.
        """
        lkey = "_%s" % key
        if lkey not in self.__dict__:
            try:
                self.__dict__[lkey] = self._loaded_configuration[key]
            except KeyError:
                if default:
                    self.__dict__[lkey] = default
                else:
                    self.__dict__[lkey] = _RuntimeConfig.DEFAULT_CONFIG[key]
                if exception:
                    raise KeyError(exception)
        return self.__dict__[lkey]

    def reset(self):
        """Resets the runtime configuration to empty state"""
        for key in list(self._loaded_configuration.keys()) + list(
            _RuntimeConfig.DEFAULT_CONFIG.keys()
        ):
            try:
                del self.__dict__[f"_{key}"]
            except KeyError:
                pass
        self._loaded_configuration = {}
        self._config_file_path = None

    @property
    def logging(self) -> dict:
        """Configuration of all subsystem loggers

        :return: A logging configuration that can be fed into
            `logging.DictConfig`.
        :rtype: dict
        """
        return self._get(
            "logging",
            exception="Sorry, no logging config in the config file",
        )

    @property
    def store_uri(self) -> str:
        """URI to the store database for results

        This must be any standards-compliant string in the form of
        `transport://user:password@host-or-path:port/db`. For example,
        `postgresql://myuser:mypass@localhost/arl`.

        :return: The URI string.
        :rtype: str
        """
        return self._get("store_uri")

    @property
    def executor_bus_port(self) -> int:
        """Port of the executor's messaging bus

        palaestrai needs one bus to start it all, which is managed by the
        executor. All other buses and topics can be communicated over this
        initial bus.

        :return: The bus port, default: 4242
        :rtype: int
        """
        return self._get("executor_bus_port")

    @property
    def public_bind(self) -> bool:
        """Indicates whether to bind to all public adresses or to localhost

        This configuration setting allows the Executor and all other message
        buses to bind to all public IP addresses if set to `True`. If not,
        the buses will bind to `localhost` only.

        :return: Whether to bind to all available IP adresses (`True`) or not.
        :rtype: bool
        """
        return self._get("public_bind")

    @property
    def profile(self) -> bool:
        """Whether to enable profiling or not."""
        return self._get("profile")

    @property
    def major_domo_client_timeout(self):
        """Timeout used by the MajorDomoClient

        Returns
        -------
        int
            The timeout, in seconds
        """
        return self._get("major_domo_client_timeout")

    @property
    def major_domo_client_retries(self) -> int:
        """The number of retries the MajorDomoClient will try

        :return: the number of retries
        :rtype: int
        """
        return self._get("major_domo_client_retries")

    @property
    def logger_port(self) -> int:
        """Destination port the internal log server should use

        All spawned submodules of palaestrAI communicate their log messages
        back to the main process. This log message receiver binds to the given
        port.

        Returns
        -------
        int
            The port of the internal log server
        """
        return self._get("logger_port")

    def load(self, stream_or_dict: Union[dict, TextIO, None] = None):
        """Loads the configuration from an external source.

        The runtime configuration is initialized from the default configuration
        in ::`_RuntimeConfig.DEFAULT_CONFIG`. This method then iterates through
        the list in ::`_RuntimeConfig.CONFIG_FILE_PATHS`, subsequently updating
        the existing configuration with new values found. Finally, the given
        ::`stream_or_dict` parameter is used if present, ultimately taking
        preference over all other values.

        That means that each config file can contain only a portion of the
        overall configuration; it gets updated subsequently.

        Parameters
        ----------
        stream_or_dict : Union[dict, TextIO, None]
            Loads the runtime configuration directly from a dictionary or as
            YAML-encoded stream. If no stream is given, the default files in
            ::`.CONFIG_FILE_PATHS` will be tried as described.
        """

        if not isinstance(self._loaded_configuration, dict):
            self._loaded_configuration = {}
        has_seen_nondefault_config = False
        self._loaded_configuration.update(_RuntimeConfig.DEFAULT_CONFIG)

        yml = ruamel.yaml.YAML(typ="safe")

        for file in _RuntimeConfig.CONFIG_FILE_PATHS:
            try:
                LOG.debug("Trying to open configuration file: %s", file)
                with open(file, "r") as fp:
                    deserialized = yml.load(fp)
                    if not isinstance(deserialized, dict):
                        LOG.warning(
                            "The contents of %s could not be deserialized "
                            "to dict, skipping it.",
                            file,
                        )
                        continue
                    self._loaded_configuration.update(deserialized)
                    self._config_file_path = file
                    has_seen_nondefault_config = True
            except IOError:
                continue

        if isinstance(stream_or_dict, dict):
            self._loaded_configuration.update(stream_or_dict)
            self._config_file_path = "(dict)"
        elif stream_or_dict is not None:
            try:
                deserialized = yml.load(stream_or_dict)  # Can raise
                if not isinstance(deserialized, dict):
                    raise TypeError
                self._loaded_configuration.update(deserialized)
                self._config_file_path = stream_or_dict
                has_seen_nondefault_config = True
            except TypeError:
                LOG.warning(
                    "Failed to load runtime configuration from stream "
                    'at "%s", ignoring.',
                    repr(stream_or_dict),
                )

        if not has_seen_nondefault_config:
            LOG.info(
                "No runtime configuration given, loaded built-in defaults."
            )
            self._config_file_path = "(DEFAULT)"

    def to_dict(self):
        return {key: self._get(key) for key in _RuntimeConfig.DEFAULT_CONFIG}

    def __str__(self):
        return "<RuntimeConfig id=0x%x> at %s" % (
            id(self),
            self._config_file_path,
        )

    def __repr__(self):
        return str(self.to_dict())


def RuntimeConfig():
    if _RuntimeConfig._instance is None:
        _RuntimeConfig._instance = _RuntimeConfig()
        try:
            _RuntimeConfig._instance.load()
        except FileNotFoundError:
            from copy import deepcopy

            _RuntimeConfig._instance._loaded_configuration = deepcopy(
                _RuntimeConfig.DEFAULT_CONFIG
            )
    return _RuntimeConfig._instance



import unittest
from unittest.mock import AsyncMock, MagicMock, patch
from uuid import uuid4

from palaestrai.agent.actuator_information import ActuatorInformation
from palaestrai.agent.agent_conductor import AgentConductor
from palaestrai.agent.dummy_brain import DummyBrain
from palaestrai.agent.sensor_information import SensorInformation
from palaestrai.core.protocol import (
    AgentSetupRequest,
    AgentSetupResponse,
    ShutdownRequest,
)


class TestAgentConductor(unittest.IsolatedAsyncioTestCase):
    def setUp(self):

        self.agent_params = {
            "name": "defender",
            "brain": {
                "name": "palaestrai.agent.dummy_brain:DummyBrain",
                "params": {"muscle_connection": "test_con"},
            },
            "muscle": {
                "name": "palaestrai.agent.dummy_muscle:DummyMuscle",
                "params": {
                    "broker_connection": "test_con",
                    "brain_connection": "test_con",
                },
            },
            "objective": {
                "name": "palaestrai.agent.dummy_objective:DummyObjective",
                "params": {"params": 1},
            },
            "sensors": [SensorInformation(0, MagicMock(), "TestSensor-1")],
            "actuators": [
                ActuatorInformation(0, MagicMock(), "TestActuator-1")
            ],
        }

        self.ac = AgentConductor(
            "test_conn", self.agent_params, 0, str(uuid4())
        )
        self.setup_req = AgentSetupRequest(
            receiver_agent_conductor=self.ac.uid,
            sender_simulation_controller="0",
            experiment_run_id="1",
            agent_id="2",
            sensors=[
                SensorInformation(0, MagicMock(), "TestSensor-1"),
                SensorInformation(0, MagicMock(), "TestSensor-2"),
            ],
            actuators=[
                ActuatorInformation(0, MagicMock(), "TestActuator-1"),
                ActuatorInformation(0, MagicMock(), "TestActuator-2"),
            ],
            agent_name="TestAgent",
        )
        self.setup_req_empty = AgentSetupRequest(
            receiver_agent_conductor=self.ac.uid,
            sender_simulation_controller="0",
            experiment_run_id="1",
            agent_id="2",
            sensors=list(),
            actuators=list(),
            agent_name="TestAgent",
        )
        self.shutdown_req = ShutdownRequest("1")

    @patch("palaestrai.agent.agent_conductor.aiomultiprocess.Process")
    def test_init_brain(self, mockaio):
        self.ac._init_brain(self.setup_req.sensors, self.setup_req.actuators)

        self.assertEqual(mockaio.call_count, 1)
        self.assertIsInstance(self.ac.brain, DummyBrain)
        self.assertEqual(len(self.ac.brain.sensors), 2)
        self.assertEqual(len(self.ac.brain.actuators), 2)

    @patch("palaestrai.agent.agent_conductor.aiomultiprocess.Process")
    def test_init_muscle(self, mockaio):
        self.ac._init_muscle(uuid4())

        self.assertEqual(mockaio.call_count, 1)
        self.assertEqual(len(self.ac.agents), 1)
        self.assertEqual(len(self.ac.tasks), 1)

    @patch("palaestrai.agent.agent_conductor.aiomultiprocess.Process")
    def test_handle_agent_setup(self, mockaio):
        self.ac._init_brain = MagicMock()
        self.ac._init_muscle = MagicMock()

        rsp = self.ac._handle_agent_setup(self.setup_req)
        self.ac._init_brain.assert_called_once()
        self.ac._init_muscle.assert_called()
        self.assertIsInstance(rsp, AgentSetupResponse)

    @patch("palaestrai.agent.agent_conductor.aiomultiprocess.Process")
    async def test_handle_shutdown(self, mockaio):
        self.ac.brain_process = AsyncMock()
        self.ac.tasks.append(AsyncMock())
        await self.ac._handle_shutdown(self.shutdown_req)

        for task in self.ac.tasks:
            self.assertEqual(task.join.call_count, 1)
        self.assertEqual(self.ac.brain_process.join.call_count, 1)

    @patch(f"{AgentConductor.__module__}.asyncio")
    async def test_housekeeping_setup(self, mock_asyncio):
        mock_transceive_task = AsyncMock()
        mock_transceive_task.result = MagicMock(
            return_value=AgentSetupRequest(
                "sim-1", "ac-1", "run-1", "ag-1", list(), list(), "agent"
            )
        )

        mock_asyncio.wait = AsyncMock(
            return_value=([mock_transceive_task], list())
        )
        mock_asyncio.create_task = MagicMock(return_value=mock_transceive_task)
        self.ac._worker = MagicMock()

        request = await self.ac._housekeeping(None)

        self.assertIsInstance(request, AgentSetupRequest)
        mock_asyncio.create_task.assert_called_once()
        mock_asyncio.wait.assert_called_once()

    @patch(f"{AgentConductor.__module__}.asyncio")
    async def test_housekeeping_shutdown(self, mock_asyncio):
        mock_transceive_task = AsyncMock()
        mock_transceive_task.result = MagicMock(
            return_value=ShutdownRequest("run-1")
        )
        mock_muscle_task = AsyncMock()
        mock_muscle_task.join = MagicMock()
        mock_muscle_task.exitcode.return_value = 0
        mock_muscle_task.is_alive = MagicMock(return_value=True)
        mock_brain_task = AsyncMock()
        mock_brain_task.join = MagicMock()
        mock_brain_task.exitcode.return_value = 0
        mock_brain_task.is_alive = MagicMock(return_value=True)

        self.ac.tasks.append(mock_muscle_task)
        self.ac.brain_process = mock_brain_task

        mock_asyncio.wait = AsyncMock(
            return_value=([mock_transceive_task], list())
        )
        mock_asyncio.create_task = MagicMock(return_value=mock_transceive_task)
        self.ac._worker = MagicMock()

        request = await self.ac._housekeeping(None)

        self.assertIsInstance(request, ShutdownRequest)
        self.assertEqual(3, mock_asyncio.create_task.call_count)
        mock_asyncio.wait.assert_called_once()

    @patch(f"{AgentConductor.__module__}.asyncio")
    async def test_housekeeping_ungraceful_death(self, mock_asyncio):
        mock_transceive_task = AsyncMock()
        mock_transceive_task.result = MagicMock(
            return_value=ShutdownRequest("run-1")
        )
        mock_muscle_task = AsyncMock()
        mock_muscle_task.join = MagicMock()
        mock_muscle_task.exitcode.return_value = 1
        mock_muscle_task.is_alive = MagicMock(return_value=False)
        mock_brain_task = AsyncMock()
        mock_brain_task.join = MagicMock()
        mock_brain_task.exitcode.return_value = 0
        mock_brain_task.is_alive = MagicMock(return_value=True)

        self.ac.tasks.append(mock_muscle_task)
        self.ac.brain_process = mock_brain_task

        mock_asyncio.wait = AsyncMock(
            return_value=([mock_transceive_task], list())
        )
        mock_asyncio.create_task = MagicMock(return_value=mock_transceive_task)
        self.ac._worker = MagicMock()

        with self.assertRaises(RuntimeError):
            await self.ac._housekeeping(None)

        # self.assertIsInstance(request, ShutdownRequest)
        self.assertEqual(3, mock_asyncio.create_task.call_count)
        mock_asyncio.wait.assert_called_once()

    @patch(f"{AgentConductor.__module__}.asyncio")
    async def test_housekeeping_signal_received(self, mock_asyncio):
        def raise_system_exit():
            raise SystemExit()

        mock_transceive_task = AsyncMock()
        mock_transceive_task.result = MagicMock(
            return_value=ShutdownRequest("run-1")
        )
        mock_muscle_task = AsyncMock()
        mock_muscle_task.join = MagicMock()
        mock_muscle_task.exitcode.return_value = 0
        mock_muscle_task.is_alive = MagicMock(return_value=True)
        mock_brain_task = AsyncMock()
        mock_brain_task.join = MagicMock()
        mock_brain_task.exitcode.return_value = 0
        mock_brain_task.is_alive = MagicMock(return_value=True)

        self.ac.tasks.append(mock_muscle_task)
        self.ac.brain_process = mock_brain_task

        mock_asyncio.wait = AsyncMock(
            side_effect=lambda x, return_when: raise_system_exit()
        )
        mock_asyncio.create_task = MagicMock(return_value=mock_transceive_task)
        self.ac._worker = MagicMock()

        with self.assertRaises(SystemExit):
            await self.ac._housekeeping(None)

        self.assertEqual(3, mock_asyncio.create_task.call_count)
        mock_asyncio.wait.assert_called_once()

    @patch("palaestrai.agent.agent_conductor.aiomultiprocess.Process")
    async def test_run(self, mockaio):
        self.ac._worker = AsyncMock()
        setup_msg = self.setup_req_empty
        shutdown_msg = self.shutdown_req
        self.ac._housekeeping = AsyncMock(
            side_effect=[setup_msg, shutdown_msg, shutdown_msg]
        )

        self.ac._handle_shutdown = AsyncMock()
        await self.ac.run()

        self.assertEqual(2, self.ac._housekeeping.call_count)


if __name__ == "__main__":
    unittest.main()



class RequestIsNoneError(Exception):
    """Raised when a request is None."""


class ReplyIsNoneError(Exception):
    """Raised when a reply is None."""


class InvalidRequestError(Exception):
    """Raised when a wrong request was received."""

    def __init__(self, expected, got):
        super().__init__(f"Expected {expected}, got {got}.")


class InvalidResponseError(Exception):
    """Raised when a wrong response was received."""

    def __init__(self, expected, got):
        super().__init__(f"Expected {expected}, got {got}.")


class TasksNotFinishedError(Exception):
    """Raised when neither the signal_monitor_task nor the
    transceive_task returned.
    """


class SignalInterruptError(Exception):
    """Raised when a signal was received."""


class InitializationFailedError(Exception):
    """Raised when initialization of an experiment failed."""


class ExperimentSetupFailedError(Exception):
    """Raised when the experiment setup failed."""


class DeadChildrenRisingAsZombiesError(Exception):
    """Raised when childrens died and starting to become zombies."""


class PrematureTaskDeathError(Exception):
    """Raised when a tasks died for its time."""


class ExperimentAlreadyRunningError(Exception):
    """Raised during experiment start if experiment is already running."""


class SimControllerFailedError(Exception):
    """Raised when a simulation controller could not be started."""


class EnvConductorFailedError(Exception):
    """Raised when an error during the execution of an environment
    conductor occurs.
    """


class AgentConductorFailedError(Exception):
    """Raised when an error during the execution of an agent conductor
    occurs.
    """


class OutOfActionSpaceError(Exception):
    """This error is raised if an :class:`.ActuatorInformation`
    receives a setpoint that is not contained in the action space
    of that actuator.
    """


class OutOfObservationSpaceError(Exception):
    """This error is raised if a :class:`.SensorInformation` is
    created with a value that is not contained in the observation
    space of that sensor.
    """


class UnknownModeError(Exception):
    """This error is raised if a :class:`.Muscle` is receiving
    an unknown Mode."""



import unittest
from unittest.mock import MagicMock, PropertyMock, patch

from palaestrai.experiment import RunGovernor
from palaestrai.experiment.rungovernor_states import (
    RGSErrorHandlingStarting,
    RGSStartingAgentConductors,
    RGSStartingEnvConductors,
)


class TestStartingSimControllers(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.rungov = RunGovernor(None, None)
        self.rgs = RGSStartingEnvConductors(self.rungov)
        self.rungov.state = self.rgs

    async def test_run(self):
        self.rgs._init_environment_conductors = MagicMock()

        await self.rgs.run()

        self.rgs._init_environment_conductors.assert_called_once()

    @patch(f"{RGSStartingEnvConductors.__module__}.aiomultiprocess.Process")
    def test_init_env_conductors(self, mock_amp):
        ec = MagicMock()
        type(ec).uid = PropertyMock(return_value="ec-1")

        env_cond = MagicMock()
        env_cond.values = MagicMock(return_value=[ec])
        self.rungov.experiment_run = MagicMock()
        self.rungov.experiment_run.environment_conductors = MagicMock(
            return_value=env_cond
        )

        self.rgs._init_environment_conductors()

        self.assertEqual(len(self.rungov.env_conductors), 1)
        self.assertIn("ec-1", self.rungov.env_conductors)

    def test_next_state(self):

        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSStartingAgentConductors)

    def test_next_state_with_errors(self):
        self.rgs.add_error(ValueError())

        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSErrorHandlingStarting)


if __name__ == "__main__":
    unittest.main()



import sqlalchemy as sa
from sqlalchemy.ext.declarative import declarative_base

metadata = sa.MetaData(
    naming_convention={
        "ix": "ix__%(column_0_label)s",
        "uq": "uq__%(table_name)s__%(column_0_name)s",
        "ck": "ck__%(table_name)s__%(column_0_name)s",
        "fk": "fk__%(table_name)s__%(column_0_name)s__%(referred_table_name)s",
        "pk": "pk__%(table_name)s",
    }
)

Base = declarative_base(metadata=metadata)



import unittest
from unittest.mock import AsyncMock, MagicMock

from palaestrai.core.protocol import (
    ExperimentRunStartRequest,
    ExperimentRunStartResponse,
)
from palaestrai.experiment import RunGovernor
from palaestrai.experiment.rungovernor_states import (
    RGSErrorHandlingStarting,
    RGSStartingSetup,
    RGSStartingSimControllers,
)
from palaestrai.util.exception import ExperimentAlreadyRunningError


class TestStartingSetup(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.rungov = RunGovernor(None, None)
        self.rgs = RGSStartingSetup(self.rungov)
        self.rungov.state = self.rgs
        self.request = ExperimentRunStartRequest(
            sender_executor_id="0",
            receiver_run_governor_id="1",
            experiment_run_id="2",
            experiment_run=None,
        )
        # TODO: Add and test the NextPhaseRequest case.

    async def test_run(self):
        self.rgs._is_experiment_run_running = MagicMock(return_value=False)
        self.rgs._experiment_run_start = AsyncMock()
        self.rungov.last_request.append(self.request)

        await self.rgs.run()

        self.rgs._is_experiment_run_running.assert_called_with(self.request)
        self.rgs._experiment_run_start.assert_called_with(self.request)

    async def test_run_experiment_already_running(self):
        self.rgs._is_experiment_run_running = MagicMock(return_value=True)
        self.rgs._experiment_start = AsyncMock()
        self.rungov.last_request.append(self.request)

        await self.rgs.run()

        self.rgs._is_experiment_run_running.assert_called_with(self.request)
        self.rgs._experiment_start.assert_not_called()

    def test_next_state(self):

        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSStartingSimControllers)

    def test_next_state_with_errors(self):
        self.rgs.add_error(ValueError())

        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSErrorHandlingStarting)

    def test_experiment_run_is_not_running(self):
        self.rungov.experiment_run = None

        res = self.rgs._is_experiment_run_running(self.request)

        self.assertFalse(res)
        self.assertEqual(0, len(self.rungov.errors))

    def test_experiment_is_running(self):
        self.rungov.experiment_run = 1

        res = self.rgs._is_experiment_run_running(self.request)

        self.assertTrue(res)
        self.assertEqual(1, len(self.rungov.errors))
        self.assertIsInstance(
            self.rungov.errors[0][0], ExperimentAlreadyRunningError
        )

    async def test_experiment_run_start(self):
        exp = MagicMock()
        self.request.experiment_run = exp

        await self.rgs._experiment_run_start(self.request)

        self.assertEqual(1, len(self.rungov.next_reply))
        self.assertIsInstance(
            self.rungov.next_reply[0], ExperimentRunStartResponse
        )

    async def test_experiment_start_with_error(self):
        exp = MagicMock()
        exp.setup = MagicMock(side_effect=[])
        self.request.experiment_run = exp

        await self.rgs._experiment_run_start(self.request)

        exp.setup.assert_called_once()
        self.assertEqual(1, len(self.rungov.errors))
        self.assertEqual(1, len(self.rungov.next_reply))
        self.assertIsInstance(
            self.rungov.next_reply[0], ExperimentRunStartResponse
        )
        self.assertFalse(self.rungov.next_reply[0].successful)


if __name__ == "__main__":
    unittest.main()



import builtins
import sys
from importlib import import_module

from . import LOG


class ErrorDuringImport(Exception):
    """Errors that occurred while trying to import something"""

    def __init__(self, filename, exc_info):
        self.filename = filename
        self.exc, self.value, self.tb = exc_info

    def __str__(self):
        exc = self.exc.__name__
        return "problem in %s - %s: %s" % (self.filename, exc, self.value)


def safeimport(path, forceload=0, cache=None):
    """Import a module; handle errors; return None if the module isn't found.

    If the module *is* found but an exception occurs, it's wrapped in an
    ErrorDuringImport exception and reraised.  Unlike __import__, if a
    package path is specified, the module at the end of the path is
    returned, not the package at the beginning.  If the optional
    'forceload' argument is 1, we reload the module from disk
    (unless it's a dynamic extension).
    """

    if cache is None:
        cache = dict()
    LOG.debug("Trying to import '%s'" % path)
    try:
        # If forceload is 1 and the module has been previously loaded from
        # disk, we always have to reload the module.  Checking the file's
        # mtime isn't good enough (e.g. the module could contain a class
        # that inherits from another module that has changed).
        if forceload and path in sys.modules:
            if path not in sys.builtin_module_names:
                # Remove the module from sys.modules and re-import to try
                # and avoid problems with partially loaded modules.
                # Also remove any submodules because they won't appear
                # in the newly loaded module's namespace if they're
                # already in sys.modules.
                subs = [m for m in sys.modules if m.startswith(path + ".")]
                for key in [path] + subs:
                    # Prevent garbage collection.
                    cache[key] = sys.modules[key]
                    del sys.modules[key]
        module = __import__(path)
    except Exception:
        # Did the error occur before or after the module was found?
        (exc, value, tb) = info = sys.exc_info()
        if path in sys.modules:
            # An error occurred while executing the imported module.
            raise ErrorDuringImport(sys.modules[path].__file__, info)
        elif exc is SyntaxError:
            # A SyntaxError occurred before we could execute the module.
            raise ErrorDuringImport(value.filename, info)
        elif exc is ImportError and value.name == path:
            # No such module in the path.
            return None
        else:
            # Some other error occurred during the importing process.
            raise ErrorDuringImport(path, sys.exc_info())
    for part in path.split(".")[1:]:
        try:
            module = getattr(module, part)
        except AttributeError:
            return None
    return module


def locate(path, forceload=0):
    """Locate an object by name or dotted path, importing as necessary."""
    parts = [part for part in path.split(".") if part]
    module, n = None, 0
    while n < len(parts) - 1:
        nextmodule = safeimport(".".join(parts[: n + 1]), forceload)
        if nextmodule:
            module, n = nextmodule, n + 1
        else:
            break
    if module:
        object = module
    else:
        object = builtins
    for part in parts[n:]:
        try:
            object = getattr(object, part)
        except AttributeError as err:
            LOG.error("Could not load '%s': %s" % (part, err))
            return None
    return object


def load_with_params(module_path, params):
    """Load a class from *module_path* and pass *params*."""
    module, clazz = module_path.split(":")
    module = import_module(module)
    return getattr(module, clazz)(**params)



"""This module contains the class :class:`ExperimentRun` that defines
an experiment run and contains all the information needed to execute
it.
"""
from __future__ import annotations

import collections.abc
from typing import TYPE_CHECKING, Dict, List, Union

from ..types.mode import Mode
from ..util import seeding
from ..util.exception import UnknownModeError

if TYPE_CHECKING:
    from palaestrai.simulation import SimulationController
    from palaestrai.experiment import TerminationCondition

import uuid

import pkg_resources
import ruamel.yaml as yml
from palaestrai.agent import AgentConductor
from palaestrai.environment import EnvironmentConductor
from palaestrai.util.dynaloader import load_with_params
from ruamel.yaml.constructor import ConstructorError

from . import LOG


class RunDefinitionError(RuntimeError):
    def __init__(self, run: ExperimentRun, message):
        super().__init__(message)
        self.message = message
        self.run = run

    def __str__(self):
        return "%s (%s)" % (self.message, self.run)


class ExperimentRun:
    """Defines an experiment run and stores information.

    The experiment run class defines a run in palaestrAI. It contains
    all information needed to execute the run. With the setup function
    the experiment run can be build.

    Parameters
    ----------

    """

    def __init__(
        self,
        user_id: Union[str, None],
        seed: Union[int, None],
        version: Union[str, None],
        schedule: List[Dict],
        run_config: dict,
    ):
        if seed is None:
            # numpy expects a seed between 0 and 2**32 - 1
            self.seed: int = seeding.create_seed(max_bytes=4)
        else:
            self.seed = seed

        if user_id is None:
            self.user_id = f"ExperimentRun-{uuid.uuid4()}"
            LOG.warning(
                "Experiment run has no user_id, please set one to "
                "identify it (assign the 'user_id' key). Generated: "
                "'%s', so that you can find it in the store.",
                self.user_id,
            )
        else:
            self.user_id = user_id

        palaestrai_version = pkg_resources.require("palaestrai")[0].version
        if version is None:
            self.version = palaestrai_version
            LOG.warning(
                "No version has been specified. There is no guarantee "
                "that this run will be executed without errors. Please "
                "set the version (assign the 'version' key) in the run "
                "file. Current palaestrAI version is '%s'.",
                self.version,
            )
        elif version != palaestrai_version:
            self.version = version
            LOG.warning(
                "Your palaestrAI installation has version %s but your "
                "run file uses version %s, which may be incompatible.",
                palaestrai_version,
                version,
            )
        else:
            self.version = version

        self.schedule_config = schedule
        self.run_config = run_config
        self.run_governor_termination_condition: TerminationCondition
        self.schedule: list

    def setup(self, broker_uri):
        """Set up an experiment run.

        Creates and configures relevant actors.
        """
        LOG.debug(
            "ExperimentRun(id=0x%x, user_id=%s) setup.", id(self), self.user_id
        )
        rgtc = self.run_config["condition"]
        LOG.debug(
            "ExperimentRun(id=0x%x, user_id=%s) loading RunGovernor "
            "TerminationCondition: %s.",
            id(self),
            self.user_id,
            rgtc["name"],
        )
        try:
            rgtc = load_with_params(rgtc["name"], rgtc["params"])
        except Exception as err:
            LOG.critical(
                "Could not load termination condition '%s' with params "
                "%s for RunGovernor: %s",
                rgtc["name"],
                rgtc["params"],
                err,
            )
            raise err
        self.run_governor_termination_condition = rgtc

        self.schedule = list()
        config = dict()
        for num, phase in enumerate(self.schedule_config):
            if len(phase) > 1:
                raise RunDefinitionError(
                    self,
                    (
                        "Only one phase per phase allowed but "
                        f"found {len(phase)} phases."
                    ),
                )
            elif len(phase) < 1:
                LOG.warning(
                    "ExperimentRun(id=0x%x, user_id=%s) found empty phase: "
                    "%d, skipping this one.",
                    id(self),
                    self.user_id,
                    num,
                )
                continue
            phase_name = list(phase.keys())[0]
            config = update_dict(config, phase[phase_name])
            agent_configs = dict()

            self.schedule.append(dict())
            self.schedule[num]["phase_config"] = config["phase_config"].copy()
            for env_config in config["environments"]:
                self.schedule[num].setdefault("environment_conductors", dict())

                ec = EnvironmentConductor(
                    env_config["environment"],
                    env_config["reward"],
                    broker_uri,
                    self.seed,
                )
                self.schedule[num]["environment_conductors"][ec.uid] = ec
            LOG.debug(
                "ExperimentRun(id=0x%x, uid=%s) set up %d "
                "EnvironmentConductor object(s) for phase %d: '%s'",
                id(self),
                self.user_id,
                len(self.schedule[num]["environment_conductors"]),
                num,
                phase_name,
            )
            if len(self.schedule[num]["environment_conductors"]) == 0:
                raise RunDefinitionError(
                    self, f"No environments defined for phase {num}."
                )

            for agent_config in config["agents"]:
                self.schedule[num].setdefault("agent_conductors", dict())

                ac_conf = {key: value for key, value in agent_config.items()}
                ac = AgentConductor(
                    broker_uri, ac_conf, self.seed, phase_name, str(id(self))
                )
                self.schedule[num]["agent_conductors"][ac.uid] = ac
                agent_configs[ac.uid] = ac_conf

            LOG.debug(
                "ExperimentRun(id=0x%x, user_id=%s) set up %d AgentConductor "
                "object(s) for phase %d: '%s'.",
                id(self),
                self.user_id,
                len(self.schedule[num]["agent_conductors"]),
                num,
                phase_name,
            )
            if len(self.schedule[num]["agent_conductors"]) == 0:
                raise RunDefinitionError(
                    self, f"No agents defined for phase {num}."
                )

            for _ in range(int(config["phase_config"].get("worker", 1))):
                self.schedule[num].setdefault("simulation_controllers", dict())
                try:
                    mode = Mode[
                        config["phase_config"].get("mode", "train").upper()
                    ]
                except KeyError as err:
                    raise UnknownModeError(err)

                sc: SimulationController = load_with_params(
                    config["simulation"]["name"],
                    {
                        "sim_connection": broker_uri,
                        "rungov_connection": broker_uri,
                        "agent_conductor_ids": self.schedule[num][
                            "agent_conductors"
                        ].keys(),
                        "environment_conductor_ids": self.schedule[num][
                            "environment_conductors"
                        ].keys(),
                        "termination_conditions": config["simulation"][
                            "conditions"
                        ],
                        "agents": agent_configs,
                        "mode": mode,
                    },
                )
                self.schedule[num]["simulation_controllers"][sc.uid] = sc
            LOG.debug(
                "ExperimentRun(id=0x%x, user_id=%s) set up %d "
                "SimulationController object(s) for phase %d: '%s'.",
                id(self),
                self.user_id,
                len(self.schedule[num]["simulation_controllers"]),
                num,
                phase_name,
            )
            if len(self.schedule[num]["simulation_controllers"]) == 0:
                raise RunDefinitionError(
                    self,
                    "No simulation controller defined. Either "
                    "'workers' < 1 or 'name' of key 'simulation' is "
                    "not available.",
                )
        LOG.info(
            "ExperimentRun(id=0x%x, user_id=%s) setup complete.",
            id(self),
            self.user_id,
        )

    def environment_conductors(self, phase=0):
        return self.schedule[phase]["environment_conductors"]

    def agent_conductors(self, phase=0):
        return self.schedule[phase]["agent_conductors"]

    def simulation_controllers(self, phase=0):
        return self.schedule[phase]["simulation_controllers"]

    def get_phase_name(self, phase: int):
        return list(self.schedule_config[phase].keys())[0]

    def get_episodes(self, phase: int):
        return self.schedule[phase]["phase_config"].get("episodes", 1)

    @property
    def num_phases(self):
        """The number of phases in this experiment run's schedule."""
        return len(self.schedule)

    def has_next_phase(self, current_phase):
        """Return if this run has a subsequent phase.

        Parameters
        ----------
        current_phase: int
            Index of the phase that is being executed.

        Returns
        -------
        bool
            True if at least one phase is taking place after
            the current phase.
        """
        return current_phase + 1 < self.num_phases

    @staticmethod
    def load(stream):
        LOG.debug("Loading configuration from %s.", stream)

        if isinstance(stream, str):
            try:
                stream = open(stream, "r")
            except OSError as err:
                LOG.error("Could not open run configuration: %s.", err)
                raise err
        try:
            conf = yml.YAML(typ="safe", pure=True).load(stream)
        except ConstructorError as err:
            LOG.error("Could not load run configuration: %s.", err)
            raise err
        finally:
            stream.close()
        LOG.debug("Loaded configuration: %s.", conf)

        return ExperimentRun(
            user_id=conf.get("user_id", conf.get("id", None)),
            seed=conf.get("seed", None),
            version=conf.get("version", None),
            schedule=conf["schedule"],
            run_config=conf["run_config"],
        )


def update_dict(src, upd):
    """Recursive update of dictionaries.

    See stackoverflow:

        https://stackoverflow.com/questions/3232943/
        update-value-of-a-nested-dictionary-of-varying-depth

    """
    for key, val in upd.items():
        if isinstance(val, collections.abc.Mapping):
            src[key] = update_dict(src.get(key, {}), val)
        else:
            src[key] = val
    return src



"""This module contains the abstract class :class:`Brain` that is used
to implement the thinking part of agents.

"""
from __future__ import annotations

import asyncio
import logging
import pickle
import queue
import signal
import time
import zlib
from abc import ABC, abstractmethod
from threading import Thread
from typing import TYPE_CHECKING, List

import setproctitle
import zmq

from .state import State
from ..core.protocol import (
    MuscleUpdateRequest,
    MuscleUpdateResponse,
    MuscleShutdownRequest,
    MuscleShutdownResponse,
)

LOG = logging.getLogger(__name__)

if TYPE_CHECKING:
    from . import SensorInformation, ActuatorInformation, Objective


class Brain(ABC):
    """Baseclass for all brain implementation

    The brain is the central learning instance. It coordinates all
    muscles (if multiple muscles are available). The brain does all
    (deep) learning tasks and delivers a model to the muscles.

    The brain has one abstract method :meth:`.thinking` that has to be
    implemented.

    Parameters
    ----------
    muscle_connection : str
        The IP and port on which the brain should bind, so that the
        muscles can connect to the brain, e.g: 'tcp://127.0.0.1:1234'.
    sensors : list[SensorInformation]
        A *list* of available sensors, can be used to define, e.g., the
        input space of a neural network.
    actuators : list[ActuatorInformation]
        A *list* of available actuators, can be used to define, e.g.,
        the output space of a neural network.
    max_buffer : int
        Size of the queue buffer.

    """

    def __init__(
        self,
        muscle_connection: str,
        sensors: List["SensorInformation"],
        actuators: List["ActuatorInformation"],
        objective: Objective,
        store_path: str,
        seed: int,
        max_buffer: int = 0,
        **params,
    ):
        self._state = State.PRISTINE
        self.seed = seed
        self.muscle_connection = muscle_connection
        self._worker_updates: queue.Queue
        self.max_buffer = max_buffer
        self.sensors = sensors
        self.actuators = actuators
        self.objective = objective
        self.store_path = store_path
        self._ctx = None
        self._router_socket = None

    @property
    def state(self) -> State:
        return self._state

    @property
    def context(self):
        if self._ctx is None:
            self._ctx = zmq.Context()
        return self._ctx

    @property
    def worker_updates(self):
        if (
            not hasattr(self, "_worker_updates")
            or self._worker_updates is None
        ):
            self._worker_updates = queue.Queue(self.max_buffer)
        return self._worker_updates

    def _handle_sigintterm(self, signum, frame):
        self._state = State.CANCELLED
        LOG.warning(
            "Brain(id=0x%x) interrupted by signal %s in frame %s.",
            id(self),
            signum,
            frame,
        )

    def _receive_updates(self):
        """Receives updates from workers and stores them in the queue."""
        self._connect()
        while self.state == State.RUNNING:
            try:
                z = self._router_socket.recv_multipart(flags=zmq.NOBLOCK)
            except zmq.ZMQError:
                time.sleep(0.05)
                continue
            p = zlib.decompress(z[1])
            msg = pickle.loads(p)
            if msg is not None:
                self.worker_updates.put([msg, z[0]])
            if isinstance(msg, MuscleShutdownRequest):
                self._state = State.STOPPING
        LOG.debug("Brain(id=0x%x) has stopped receiver.", id(self))

    def _connect(self):
        """connect the core channels to the specified uris"""
        self._router_socket = self.context.socket(zmq.ROUTER)
        self._router_socket.bind(self.muscle_connection)
        LOG.debug(
            "Brain(id=0x%x) bound to %s", id(self), self.muscle_connection
        )

    def _send(self, message, header, flags=0, protocol=-1):
        """Send a message to a muscle.

        Parameters
        ----------
        message : MuscleUpdateResponse
            The message that should be send as response.
        header : str or bytes
            The ID of the receiving muscle.
        flags : int, optional
            ZeroMQ-Flags which should be used when sending.
        protocol : int, optional
            Protocol used by pickle.dumps

        """
        p = pickle.dumps(message, protocol)
        z = zlib.compress(p)
        self._router_socket.send_multipart([bytes(header), z], flags=flags)

    async def run(self):
        """Start the brain main loop.

        This method starts the brain, it begins to listen for messages
        and as soon as messages are arriving, it processes those
        messages by calling the thinking methode. This will return a
        Muscleupdateresponse, which is sent back to the muscle.
        """
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        signal.signal(signal.SIGABRT, signal.SIG_DFL)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        setproctitle.setproctitle("palaestrAI[Brain-%s]")

        self._state = State.RUNNING
        signal.signal(signal.SIGINT, self._handle_sigintterm)
        signal.signal(signal.SIGTERM, self._handle_sigintterm)
        receiver_thread = Thread(target=self._receive_updates)
        receiver_thread.start()
        LOG.info(
            "Brain(id=0x%x) started: complex ideas will now become real.",
            id(self),
        )

        while self.state == State.RUNNING or not self.worker_updates.empty():
            try:
                msg = self.worker_updates.get(timeout=1)
            except TimeoutError:
                continue
            except queue.Empty:
                await asyncio.sleep(0.05)
                continue
            msg, header = msg[0], msg[1]
            LOG.debug(
                "Brain(id=0x%x) received request: %s.",
                id(self),
                msg,
            )
            if isinstance(msg, MuscleUpdateRequest):
                LOG.debug(
                    "Brain(id=x%x) will think about that breaking new "
                    "MuscleUpdate that just arrived.",
                    id(self),
                )
                response = self.thinking(
                    header,
                    msg.sensor_readings,
                    msg.last_actions,
                    self.objective.internal_reward(msg.reward),
                    msg.is_terminal,
                )
                self.store_model(self.store_path)
            elif isinstance(msg, MuscleShutdownRequest):
                LOG.info(
                    "Brain(id=0x%x) saw its only muscle requesting a break.",
                    id(self),
                )
                response = MuscleShutdownResponse()
                self._state = State.STOPPING
            else:
                LOG.error(
                    "Brain(id=0x%x) "
                    "has received a message of type %s, but cannot handle it; "
                    "ignoring",
                    id(self),
                    type(msg),
                )
                continue

            self._send(response, header)

        LOG.debug(
            "Brain(id=0x%x) runner has ended, joining receiver thread",
            id(self),
        )
        receiver_thread.join()
        if self.state == State.STOPPING:
            self._state = State.FINISHED
        LOG.info("Brain(id=0x%x) completed shutdown.", id(self))

    @abstractmethod
    def thinking(
        self, muscle_id, readings, actions, reward, done
    ) -> MuscleUpdateResponse:
        """Think about a response using the provided information.

        The :meth:`.thinking` method is the place for the
        implementation of the agent's/brain's logic. The brain can
        use the current sensor readings, review the actions of the
        previous thinking and consider the reward (provided by the
        objective).

        Usually, this is the place where machine learning happens,
        but other solutions are possible as well (like a set of rules
        or even random based results).

        Parameters
        -------
        muscle_id : UID
            This is the ID of the muscle which requested the update
        readings : list[sensor_information]
            A list containing the new sensor readings
        actions : list[actuator_information]
            A list containing the actions of the last iteration
        reward : list[floats] or list[int] or float
            A float of the environment reward, a list of floats if
            multiple environments are used in parallel
        done : bool
            A boolean which signals if the simulation run has
            terminated

        Returns
        -------
        MuscleUpdateResponse
        """
        pass

    @abstractmethod
    def store_model(self, path):
        """Storing a trained Model

        After each training the model should be stored so it can
        be loaded afterwards.
        """

        pass

    @abstractmethod
    def load_model(self, path):
        """Load a previously trained model.

        A model that was trained and stored should be loaded with this
        method so it can be reused in another training session or a
        testing run.

        """
        pass



import logging

LOG = logging.getLogger(__name__)

from .simulation_controller import SimulationController
from .vanilla_sim_controller import VanillaSimController
from .vanilla_simcontroller_termination_condition import (
    VanillaSimControllerTerminationCondition,
)



from palaestrai.core.protocol import MuscleUpdateResponse

from .brain import Brain


class DummyBrain(Brain):
    def __init__(
        self,
        muscle_connection: str,
        sensors,
        actuators,
        objective,
        store_path,
        seed: int,
        **params,
    ):
        super().__init__(
            muscle_connection,
            sensors,
            actuators,
            objective,
            store_path,
            seed,
            **params,
        )

    def thinking(self, muscle_id, readings, actions, reward, done):
        response = MuscleUpdateResponse(False, None)
        return response

    def store_model(self, path):
        pass

    def load_model(self, path):
        pass



import json
import logging
import os
import queue
import threading
import time
import uuid
import zlib
from io import StringIO

import jsonpickle
import jsonpickle.ext.numpy as jsonpickle_numpy
import palaestrai.core.MDP as MDP
import palaestrai.core.protocol as proto
import ruamel.yaml as yml
import sqlalchemy
import sqlalchemy.engine
import sqlalchemy.exc
import sqlalchemy.orm
from palaestrai.core import RuntimeConfig
from palaestrai.core.serialisation import deserialize

from . import LOG
from . import database_model as dbm


class StoreReceiver(threading.Thread):
    """The message receiver of the palaestrAI store.

    The store hooks into the global communication, reading every message that
    is being exchanged between :class:`Executor`, :class:`RunGovernor`,
    :class:`AgentConductor`, :class:`Environment`, :class:`Brain`, and
    :class:`Muscle` instances. From these messages, it reads all relevant
    status informtion in order to relay them to the store database for later
    analysis of experiments.
    """

    def __init__(self, queue, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._queue = queue
        self._running = True
        self._uid = uuid.uuid4()
        self._db_engine = None
        self._db_session_maker = None
        self._db_session = None
        self._log_store_fh = None
        self._worlds_state_sensors = []  # Which sensors carry the world state
        # we know from the experiment object.
        self._message_dispatch = {
            proto.ExperimentRunStartRequest: self._write_experiment,
            proto.ExperimentRunStartResponse: None,
            proto.ExperimentRunShutdownRequest: None,
            proto.ExperimentRunShutdownResponse: None,
            proto.SimulationStartRequest: self._write_simulation_instance,
            proto.SimulationStartResponse: None,
            proto.SimulationControllerTerminationRequest: None,
            proto.SimulationControllerTerminationResponse: None,
            proto.SimulationStopRequest: None,
            proto.SimulationStopResponse: None,
            proto.EnvironmentSetupRequest: None,
            proto.EnvironmentSetupResponse: self._write_environment,
            proto.EnvironmentStartRequest: None,
            proto.EnvironmentStartResponse: None,  # self._write_world_state?
            proto.EnvironmentResetRequest: None,
            proto.EnvironmentResetResponse: None,
            proto.EnvironmentResetNotificationRequest: None,
            proto.EnvironmentResetNotificationResponse: None,
            proto.EnvironmentShutdownRequest: None,
            proto.EnvironmentShutdownResponse: None,
            proto.EnvironmentUpdateRequest: None,
            proto.EnvironmentUpdateResponse: self._write_world_state,
            proto.AgentSetupRequest: self._write_agent,
            proto.AgentSetupResponse: None,
            proto.AgentUpdateRequest: None,
            proto.AgentUpdateResponse: self._write_agent_update,
            proto.AgentShutdownRequest: None,
            proto.AgentShutdownResponse: None,
            proto.MuscleUpdateRequest: None,
            proto.MuscleUpdateResponse: None,  # We need to dump brain data
            proto.ShutdownRequest: None,
            proto.ShutdownResponse: None,
            proto.NextPhaseRequest: None,
            proto.NextPhaseResponse: None,
        }
        LOG.debug("New store Receiver instance %s ready", self._uid)
        try:
            self._store_uri = RuntimeConfig().store_uri
            if not self._store_uri:
                raise KeyError
        except KeyError:
            LOG.error(
                "StoreReceiver(id=%0xd, uid=%s) "
                "has no store_uri configured, I'm going to disable myself. :-("
                " If you want to employ me, set the 'store_uri' runtime "
                "configuration parameter.",
                id(self),
                self._uid,
            )
            self.disable()
        jsonpickle_numpy.register_handlers()

    def __del__(self):
        self.disable()

    def disable(self):
        """Disables the store completely."""
        for k in self._message_dispatch.keys():  # Disable all handlers.
            self._message_dispatch[k] = None
        if self._db_session:
            self._db_session.flush()
            self._db_session.close()
            self._db_session = None
        if self._log_store_fh:
            self._log_store_fh.flush()
            self._log_store_fh.close()
            self._log_store_fh = None

    @property
    def _log_store(self):
        if not self._log_store_fh:
            log_store_dir = os.path.join(os.getcwd(), "store.log")
            os.makedirs(log_store_dir, exist_ok=True)
            log_store_fp = os.path.join(
                log_store_dir, "%s-%s.log" % (int(time.time()), self._uid)
            )
            self._log_store_fh = open(log_store_fp, "w")
            LOG.debug(
                "StoreReceiver(id=%0xd, uid=%s) writing message log to: %s",
                id(self),
                self._uid,
                log_store_fp,
            )
        return self._log_store_fh

    @property
    def uid(self):
        return self._uid

    @property
    def _dbh(self) -> sqlalchemy.orm.session:
        if self._db_session is None:
            self._db_engine = sqlalchemy.create_engine(
                RuntimeConfig().store_uri
            )
            self._db_session_maker = sqlalchemy.orm.sessionmaker()
            self._db_session_maker.configure(bind=self._db_engine)
            try:
                self._db_session = self._db_session_maker()
            except (
                sqlalchemy.exc.OperationalError,
                sqlalchemy.exc.ArgumentError,
            ) as e:
                LOG.error(
                    "Store could not connect: %s. We can continue, "
                    "but there won't be any data stored. Sorry.",
                    e,
                )
                self.disable()
        return self._db_session

    def run(self):
        """Run the store."""
        LOG.debug(
            "Starting StoreReceiver(id=0x%x, uid=%s)", id(self), self._uid
        )
        while self._running or not self._queue.empty():
            try:
                msg = self._queue.get(timeout=1)
            except queue.Empty:
                time.sleep(1)
                continue
            msg_type, msg_uid, msg_obj = self._read(msg)
            LOG.debug(
                "StoreReceiver(id=0x%x, uid=%s) received message: "
                "type=%s, uid=%s, payload=%s",
                id(self),
                self._uid,
                msg_type,
                msg_uid,
                msg_obj,
            )
            if msg_type in ("ignore", "error"):
                continue
            if isinstance(msg_obj, list):
                LOG.info(
                    "StoreReceiver(id=0x%x, uid=%s) received a list of "
                    "%d messages. Handling all these messages separately.",
                    id(self),
                    self._uid,
                    len(msg_obj),
                )
                for msg in msg_obj:
                    self.write(msg)
            else:
                self.write(msg_obj)
        self.disable()
        LOG.debug("StoreReceiver(id=0x%x, uid=%s) closed", id(self), self._uid)

    def shutdown(self):
        LOG.debug(
            "StoreReceiver(id=0x%x, uid=%s) preparing shutdown",
            id(self),
            self._uid,
        )
        self._running = False

    def write(self, message):
        if LOG.level == logging.DEBUG:
            self._log_store.write(
                "%s StoreReceiver(id=0x%x, uid=%s)[%d] %s %s"
                % (
                    time.time(),
                    id(self),
                    self._uid,
                    os.getpid(),
                    message,
                    message.__dict__
                    if hasattr(message, "__dict__")
                    else str(message),
                )
            )
            self._log_store.flush()
        if message.__class__ not in self._message_dispatch:
            StoreReceiver._handle_unknown_message(message)
            return
        if self._message_dispatch[message.__class__] is not None:
            try:
                LOG.debug(
                    "StoreReceiver(id=0x%x, uid=%s) dispatching message %s",
                    id(self),
                    self.uid,
                    message,
                )
                self._message_dispatch[message.__class__](message)
            except (
                sqlalchemy.exc.NoForeignKeysError,
                sqlalchemy.exc.ProgrammingError,
            ) as e:
                LOG.critical(
                    "StoreReceiver(id=0x%x, uid=%s) "
                    "notes that the developers are too stupid to get the "
                    "schema right: %s",
                    id(self),
                    self.uid,
                    e,
                )
            except (
                sqlalchemy.exc.InvalidRequestError,
                sqlalchemy.exc.OperationalError,
                sqlalchemy.exc.ArgumentError,
            ) as e:
                LOG.critical(
                    "StoreReceiver(id=0x%x, uid=%s) "
                    "failed to write to the database: %s. "
                    "Please check that connecting to the database is "
                    "possible and that you have run `palaestrai "
                    "database-create'. I'm going to disable myself now. Go "
                    "on with your puny experiment, I can't keep track of it!",
                    id(self),
                    self.uid,
                    e,
                )
                self.disable()

    @staticmethod
    def _handle_unknown_message(message):
        LOG.warning(
            "Store received message %s, but cannot handle it - ignoring",
            message,
        )

    def _write_experiment(self, msg: proto.ExperimentRunStartRequest):
        from palaestrai.experiment.experiment_run import ExperimentRun

        study_hack = dbm.ExperimentStudy()  # TODO Hack, hack!!
        study_hack.name = (
            "I'm a Dummy Hacky Study :^), "
            "please replace me at %s" % __file__
        )
        self._dbh.add(study_hack)

        eq = self._dbh.query(dbm.Experiment).filter_by(
            name=msg.experiment_run.user_id
        )
        experiment_record = eq.first()
        if not experiment_record:
            sio = StringIO()
            yaml = yml.YAML(typ="safe")
            yaml.register_class(ExperimentRun)
            yaml.dump(msg.experiment_run, sio)
            experiment_record = dbm.Experiment(
                experiment_study_id=study_hack.id,
                experiment=sio.getvalue(),
                name=msg.experiment_run.user_id,
            )
            self._dbh.add(experiment_record)
            self._dbh.commit()

        experiment_run_record = (
            self._dbh.query(dbm.ExperimentRun)
            .filter_by(uuid=msg.experiment_run_id)
            .first()
        )
        if not experiment_run_record:
            experiment_run_record = dbm.ExperimentRun(
                uuid=msg.experiment_run_id, experiment_id=experiment_record.id
            )
            self._dbh.add(experiment_run_record)
            self._dbh.commit()
        else:
            LOG.warn(
                "Duplicate ExperimentRun record with uuid=%s",
                experiment_run_record.uuid,
            )
        self._worlds_state_sensors = list()
        if not isinstance(msg.experiment_run.schedule, list):
            # No list, no schedule, no world state sensors
            # no experiment? Stories from the mypy bubble...
            return
        for phase in list(msg.experiment_run.schedule):
            for env in phase["environments"]:
                for sen in env["params"]["world_state_sensors"]:
                    self._worlds_state_sensors.append(sen)

    def _write_simulation_instance(
        self, message: proto.SimulationStartRequest
    ):
        q = self._dbh.query(dbm.ExperimentRun).filter_by(
            uuid=message.experiment_run_id
        )
        experiment_run_record = q.first()
        if not experiment_run_record:
            LOG.error(
                "Store received SimulationStartRequest, but could not find "
                "run for run_id=%s",
                message.experiment_run_id,
            )
            return
        q = self._dbh.query(dbm.SimulationInstance).filter_by(
            uuid=message.simulation_controller_id
        )
        if (
            q.count() > 0
            and q.one().experiment_run_id == experiment_run_record.id
        ):
            return  # OK, we received a duplicate
        elif (
            q.count() > 0
            and q.one().experiment_run_id != experiment_run_record.id
        ):
            LOG.error(
                "StoreReceiver(id=0x%x, uid=%s) "
                "received SimulationStartRequest(simulation_controller_id=%s);"
                " the SimulationController.uid already exists, but with "
                "different experiment runs (in DB: %s, received: %s); "
                "ignoring - expect more errors",
                id(self),
                self.uid,
                message.simulation_controller_id,
                q.one().uuid,
                message.simulation_controller_id,
            )
            return  # Nothing we can do here except complaining
        LOG.debug(
            "Writing new SimulationInstance for experiment_run_id=%s",
            experiment_run_record.uuid,
        )
        self._dbh.add(
            dbm.SimulationInstance(
                uuid=message.simulation_controller_id,
                experiment_run_id=experiment_run_record.id,
            )
        )
        self._dbh.commit()

    def _write_environment(self, message: proto.EnvironmentSetupResponse):
        q = self._dbh.query(dbm.SimulationInstance).filter_by(
            uuid=message.receiver_simulation_controller
        )
        simulation_instance_record = q.first()
        self._dbh.add(
            dbm.EnvironmentConductor(
                uuid=message.sender_environment_conductor,
                type=message.environment_type,
                parameters=json.dumps(message.environment_parameters),
                simulation_instance_id=simulation_instance_record.id,
            )
        )
        self._dbh.commit()

    def _write_world_state(self, message: proto.EnvironmentUpdateResponse):

        q = self._dbh.query(dbm.EnvironmentConductor).filter_by(
            uuid=message.environment_conductor_id
        )
        environment_conductor_record = q.one()

        self._dbh.add(
            dbm.WorldState(
                environment_conductor_id=environment_conductor_record.id,
                is_terminal=message.is_terminal,
                state_dump=jsonpickle.encode(message.sensors),
            )
        )
        self._dbh.commit()

    def _write_agent(self, message: proto.AgentSetupRequest):
        q = self._dbh.query(dbm.ExperimentRun).filter_by(
            uuid=message.experiment_run_id
        )
        experiment_run_record = q.one()
        q = self._dbh.query(dbm.AgentConductor).filter_by(
            uuid=message.receiver_agent_conductor,
        )
        if q.count() == 0:
            agent_conductor_record = dbm.AgentConductor(
                experiment_run_id=experiment_run_record.id,
                uuid=message.receiver_agent_conductor,
                name=message.agent_name,
            )
            self._dbh.add(agent_conductor_record)
        else:
            agent_conductor_record = q.one()
        q = self._dbh.query(dbm.SimulationInstance).filter_by(
            uuid=message.sender_simulation_controller
        )
        simulation_instance_record = q.one()
        self._dbh.add(
            dbm.Muscle(
                agent_conductor_id=agent_conductor_record.id,
                simulation_instance_id=simulation_instance_record.id,
                uuid=message.agent_id,
            )
        )
        self._dbh.commit()

    def _write_agent_update(self, message: proto.AgentUpdateResponse):

        q = self._dbh.query(dbm.Muscle).filter_by(
            uuid=message.sender_agent_id,
        )
        muscle_record = q.one()
        muscle_action_record = dbm.MuscleAction(
            muscle_id=muscle_record.id,
            sensor_readings=jsonpickle.encode(message.sensor_information),
            actuator_information=jsonpickle.encode(
                message.actuator_information
            ),
        )
        self._dbh.add(muscle_action_record)
        self._dbh.commit()

    def _read(self, msg):
        """Unpacks a message, filters ignores"""

        _ = msg.pop(0)
        empty = msg.pop(0)
        assert empty == b""
        _ = msg.pop(0)
        # if len(msg) >= 1:
        #     serv_comm = msg.pop(0)
        if len(msg) > 3:
            sender = msg.pop(0)
            empty = msg.pop(0)
            header = msg.pop(0)
            LOG.debug(
                "Ignored message parts: %s, %s, %s", sender, empty, header
            )

        if msg[0] == MDP.W_HEARTBEAT:
            # We ignore heartbeats
            return "ignore", None, None
        if msg[0] == MDP.W_READY:
            return "ignore", None, None

        if len(msg) == 1:
            # it is a response
            uid = ""
            msg_obj = StoreReceiver._deserialize(msg.pop(0))
            msg_type = "response"

        elif len(msg) == 2:
            uid = StoreReceiver._deserialize(msg.pop(0))
            msg_obj = StoreReceiver._deserialize(msg.pop(0))
            msg_type = "request"
        else:
            print(msg)
            uid = ""
            msg_obj = "None"
            msg_type = "error"

        return msg_type, uid, msg_obj

    @staticmethod
    def _deserialize(msg):
        try:
            msg = deserialize([msg])
            return msg
        except zlib.error:
            LOG.debug("Message %s could not be deserialized.", msg)
        try:
            msg = str(msg.decode())
            return msg
        except AttributeError:
            LOG.debug("Message %s could not be decoded.", msg)

        return msg



import logging

LOG = logging.getLogger("palaestrai.experiment.run_governor")

from .base_state import RunGovernorState
from .rgs_done import RGSDone
from .rgs_errors_initializing import RGSErrorHandlingInitializing
from .rgs_errors_starting import RGSErrorHandlingStarting
from .rgs_errors_transceiving import RGSErrorHandlingTransceiving
from .rgs_finalizing import RGSFinalizing
from .rgs_handling_dead_children import RGSHandlingDeadChildren
from .rgs_handling_simctrl_termination import (
    RGSHandlingSimControllerTermination,
)
from .rgs_initializing import RGSInitializing
from .rgs_pristine import RGSPristine
from .rgs_starting_setup import RGSStartingSetup
from .rgs_starting_agent_conductors import RGSStartingAgentConductors
from .rgs_starting_env_conductors import RGSStartingEnvConductors
from .rgs_starting_run import RGSStartingRun
from .rgs_starting_sim_controllers import RGSStartingSimControllers
from .rgs_stopping_run import RGSStoppingRun
from .rgs_stopping_simulation import RGSStoppingSimulation
from .rgs_stopping_transceiving import RGSStoppingTransceiving
from .rgs_transceiving import RGSTransceiving



"""This module contains the class :class:`SensorInformation` that
stores all information the agents need about a single sensor.

"""

from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from palaestrai.types import Space


class SensorInformation:
    """Stores information about a single sensor.

    Once created, a *SensorInformation* object can be called to
    retrieve its value, e.g.,:

        a = SensorInformation(42, some_space)
        a()  # => 42
        a.sensor_value  # => 42


    Parameters
    ----------
    sensor_value: int float as described in observation_space
        The value of the sensor's last reading. The type of this value
        is described by :attr:`observation_space`
    observation_space : :class:`~.Space`
        An instance of a palaestrai space object defining the type of
        the value
    sensor_id: str or int, optional
        A unique identifier for this sensor. The agents use the ID only
        for assignment of previous values with the same ID. The ID is
        not analyzed to gain domain knowledge (e.g., if the sensor is
        called "Powergrid.Bus1", the agent will not use the ID to
        identify this sensor as part of a Bus in a powergrid.)

    """

    def __init__(self, sensor_value, observation_space: Space, sensor_id=None):
        self.sensor_value = sensor_value
        self.observation_space = observation_space
        self.sensor_id = sensor_id

    def __call__(self):
        """Reads the sensor"""
        return self.sensor_value

    def __repr__(self):
        return (
            "SensorInformation("
            f"value={self.sensor_value}, observation_space={repr(self.observation_space)}, sensor_id={self.sensor_id})"
        )

    @property
    def id(self):
        return self.sensor_id

    @id.setter
    def id(self, value):
        self.sensor_id = value



import os

_file_dir = os.path.dirname(os.path.realpath(__file__))
dummy_exp_path: str = os.path.abspath(
    f"{_file_dir}/fixtures/dummy_experiment.yml"
)
runtime_path: str = os.path.abspath(
    f"{_file_dir}/fixtures/palaestrai-runtime-debug.conf.yaml"
)



"""

"""
from palaestrai.cli.manager import cli


def test_cli_commands_exist():
    tool = cli
    for cmd in ["experiment-start", "database-create", "database-migrate"]:
        assert cmd in tool.commands



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class EnvironmentResetNotificationResponse:
    """Response to an environment reset notification.

    Parameters
    ----------
    sender_agent_id: str
        ID of the sending :class:`.Muscle`.
    receiver_simulation_controller_id: str
        ID of the receiving :class:`.SimulationController`.

    """

    sender_agent_id: str
    receiver_simulation_controller_id: str

    @property
    def sender(self):
        return self.sender_agent_id

    @property
    def receiver(self):
        return self.receiver_simulation_controller_id



import os.path
import sys
import unittest
import uuid
from unittest.mock import MagicMock, patch

import pkg_resources
from palaestrai.experiment.experiment_run import ExperimentRun


class _UidentifiableMock(MagicMock):
    def __init__(self, id=None):
        super().__init__()
        self.uid = id if id else str(uuid.uuid4())


def load_with_params_side_effect(*args, **kwargs):
    return _UidentifiableMock()


@patch(
    "palaestrai.experiment.experiment_run.AgentConductor",
    MagicMock(side_effect=load_with_params_side_effect),
)
@patch(
    "palaestrai.experiment.experiment_run.load_with_params",
    MagicMock(side_effect=load_with_params_side_effect),
)
class TestExperimentRun(unittest.TestCase):
    def setUp(self):
        self.version = pkg_resources.require("palaestrai")[0].version
        self.dummy_exp_path = os.path.realpath(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "../../../fixtures/dummy_run.yml",
            )
        )
        self.schedule = [
            {
                "phase_0": {
                    "environments": [
                        {
                            "environment": {"name": "", "params": dict()},
                            "reward": {"name": "", "params": dict()},
                        }
                    ],
                    "agents": [
                        {
                            "name": "defender",
                            "brain": {"name": "", "params": dict()},
                            "muscle": {"name": "", "params": dict()},
                            "objective": {"name": "", "params": dict()},
                            "sensors": list(),
                            "actuators": list(),
                        },
                        {
                            "name": "attacker",
                            "brain": {"name": "", "params": dict()},
                            "muscle": {"name": "", "params": dict()},
                            "objective": {"name": "", "params": dict()},
                            "sensors": list(),
                            "actuators": list(),
                        },
                    ],
                    "simulation": {
                        "name": "",
                        "params": dict(),
                        "conditions": list(),
                    },
                    "phase_config": {"mode": "train", "worker": 1},
                }
            }
        ]
        self.run_config = {
            "condition": {
                "name": (
                    "palaestrai.experiment:"
                    "VanillaRunGovernorTerminationCondition",
                ),
                "params": dict(),
            }
        }

        self.schedule_p2 = {
            "phase_1": {
                "environments": [
                    {
                        "environment": {"name": "", "params": dict()},
                        "reward": {"name": "", "params": dict()},
                    }
                ],
                "agents": [
                    {
                        "name": "defender",
                        "brain": {"name": "", "params": dict()},
                        "muscle": {"name": "", "params": dict()},
                        "objective": {"name": "", "params": dict()},
                        "sensors": list(),
                        "actuators": list(),
                    },
                ],
                "simulation": {
                    "name": "",
                    "params": dict(),
                    "conditions": list(),
                },
                "phase_config": {"mode": "test", "worker": 1},
            }
        }
        self.schedule_p3 = {
            "phase_1": {
                "environments": [
                    {
                        "environment": {"name": "", "params": dict()},
                        "reward": {"name": "", "params": dict()},
                    }
                ],
                "agents": [
                    {
                        "name": "attacker",
                        "brain": {"name": "", "params": dict()},
                        "muscle": {"name": "", "params": dict()},
                        "objective": {"name": "", "params": dict()},
                        "sensors": list(),
                        "actuators": list(),
                    },
                ],
                "simulation": {
                    "name": "",
                    "params": dict(),
                    "conditions": list(),
                },
                "phase_config": {"mode": "train", "worker": 3},
            }
        }

    def test_properties(self):
        """Assert Not Empty list"""
        exp = ExperimentRun(
            user_id="test_properties",
            seed=123,
            version=self.version,
            schedule=self.schedule,
            run_config=self.run_config,
        )
        self.assertFalse(hasattr(exp, "schedule"))

    def test_setup_one_phase(self):
        """Assert setup"""
        exp = ExperimentRun(
            user_id="test_setup",
            seed=123,
            version=self.version,
            schedule=self.schedule,
            run_config=self.run_config,
        )
        exp.setup(broker_uri=None)
        self.assertEqual(len(exp.schedule_config), len(exp.schedule))
        self.assertEqual(1, len(exp.environment_conductors(0)))
        self.assertEqual(2, len(exp.agent_conductors(0)))
        self.assertEqual(1, len(exp.simulation_controllers(0)))

    def test_setup_three_phase(self):
        self.schedule.append(self.schedule_p2)
        self.schedule.append(self.schedule_p3)
        exp = ExperimentRun(
            user_id="test_setup",
            seed=123,
            version=self.version,
            schedule=self.schedule,
            run_config=self.run_config,
        )
        exp.setup(broker_uri=None)
        self.assertEqual(len(exp.schedule_config), len(exp.schedule))
        self.assertEqual(1, len(exp.environment_conductors(0)))
        self.assertEqual(2, len(exp.agent_conductors(0)))
        self.assertEqual(1, len(exp.simulation_controllers(0)))
        self.assertEqual(1, len(exp.environment_conductors(1)))

        # One agent was removed in phase 2
        self.assertEqual(1, len(exp.agent_conductors(1)))
        self.assertEqual(1, len(exp.simulation_controllers(1)))
        self.assertEqual(1, len(exp.environment_conductors(2)))
        self.assertEqual(1, len(exp.agent_conductors(2)))

        # Now we have three workers
        self.assertEqual(3, len(exp.simulation_controllers(2)))

    def test_load_from_file(self):
        exp = ExperimentRun.load(self.dummy_exp_path)
        self.assertFalse(hasattr(exp, "schedule"))

    def test_load_from_stream(self):
        with open(self.dummy_exp_path, "r") as stream_:
            exp = ExperimentRun.load(stream_)

        self.assertFalse(hasattr(exp, "schedule"))


class TestExperimentRunDump(unittest.TestCase):
    """Needed a separate test to get rid of the patches."""

    def test_dump(self):

        from io import StringIO

        import ruamel.yaml as yml
        from palaestrai.experiment import ExperimentRun

        dummy_exp_path = os.path.realpath(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "../../../fixtures/dummy_run.yml",
            )
        )
        sio = StringIO()
        yaml = yml.YAML(typ="safe")
        yaml.register_class(ExperimentRun)
        exp = ExperimentRun.load(dummy_exp_path)
        yaml.dump(exp, sio)

        exp_loaded = yaml.load(sio.getvalue())
        self.assertEqual(sys.getsizeof(exp), sys.getsizeof(exp_loaded))


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import patch

import palaestrai.core.protocol
from palaestrai.store.receiver import StoreReceiver


class TestReceiver(unittest.TestCase):
    @patch("palaestrai.core.RuntimeConfig")
    def test_handles_all_protocol_messages(self, _):
        # Ideally, we should explicitly handle *all* message types in the
        # store receiver. I.e., everything that is defined in the module:
        # all_message_types = [
        #    t for t in palaestrai.core.protocol.__dir__()
        #    if t.endswith("Request") or t.endswith("Response")
        # ]
        # ...but until we do, we need to go with a curated list instead.
        # Simply setting all hitherto unknown types to "None" in the
        # receiver is *not* an option, as it will never indicate that there
        # are still sime types not considered.
        all_message_types = [
            palaestrai.core.protocol.ExperimentRunStartRequest,
            palaestrai.core.protocol.ExperimentRunStartResponse,
            palaestrai.core.protocol.ExperimentRunShutdownRequest,
            palaestrai.core.protocol.ExperimentRunShutdownResponse,
            palaestrai.core.protocol.SimulationStartRequest,
            palaestrai.core.protocol.SimulationStartResponse,
            palaestrai.core.protocol.SimulationControllerTerminationRequest,
            palaestrai.core.protocol.SimulationControllerTerminationResponse,
            palaestrai.core.protocol.SimulationStopRequest,
            palaestrai.core.protocol.SimulationStopResponse,
            palaestrai.core.protocol.EnvironmentSetupRequest,
            palaestrai.core.protocol.EnvironmentSetupResponse,
            palaestrai.core.protocol.EnvironmentStartRequest,
            palaestrai.core.protocol.EnvironmentStartResponse,
            palaestrai.core.protocol.EnvironmentUpdateRequest,
            palaestrai.core.protocol.EnvironmentUpdateResponse,
            palaestrai.core.protocol.EnvironmentResetRequest,
            palaestrai.core.protocol.EnvironmentResetResponse,
            palaestrai.core.protocol.EnvironmentResetNotificationRequest,
            palaestrai.core.protocol.EnvironmentResetNotificationResponse,
            palaestrai.core.protocol.EnvironmentShutdownRequest,
            palaestrai.core.protocol.EnvironmentShutdownResponse,
            palaestrai.core.protocol.AgentSetupRequest,
            palaestrai.core.protocol.AgentSetupResponse,
            palaestrai.core.protocol.AgentUpdateRequest,
            palaestrai.core.protocol.AgentUpdateResponse,
            palaestrai.core.protocol.AgentShutdownRequest,
            palaestrai.core.protocol.AgentShutdownResponse,
            palaestrai.core.protocol.MuscleUpdateRequest,
            palaestrai.core.protocol.MuscleUpdateResponse,
            palaestrai.core.protocol.ShutdownRequest,
            palaestrai.core.protocol.ShutdownResponse,
            palaestrai.core.protocol.NextPhaseRequest,
            palaestrai.core.protocol.NextPhaseResponse,
        ]
        receiver = StoreReceiver(queue=None)
        for t in all_message_types:
            try:
                _ = receiver._message_dispatch[t]
            except KeyError:
                self.fail(
                    f"Message type {t} raises key error known to the store "
                    f"dispatcher."
                )


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from palaestrai.experiment import ExperimentRun


@dataclass
class ExperimentRunStartRequest:
    """Request to a :class:`RunGovernor` to start a new experiment run

    Parameters
    ----------
    sender_executor_id: str
        ID of the sending executor.
    receiver_run_governor_id: str
        ID of the receiving run governor.
    experiment_run: :class:`ExperimentRun`
        The experiment run that should be executed.
    experiment_run_id: str
        ID for this instance run of the given run.
    """

    sender_executor_id: str
    receiver_run_governor_id: str
    experiment_run_id: str
    experiment_run: ExperimentRun

    @property
    def sender(self):
        """ID of the sending executor."""
        return self.sender_executor_id

    @property
    def receiver(self):
        """ID of the receiving run governor."""
        return self.receiver_run_governor_id

    def __str__(self):
        return (
            "ExperimentRunStartRequest for Run='%s'"
            % self.experiment_run.user_id
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .. import RunGovernor


class RunGovernorState(ABC):
    """The base class for all run governor states.

    Parameters
    ----------
    rgc: :class:`.RunGovernor`
        The run governor that provides the context for this state.
        Therefore, the abbreviation rgc (run governor context)
    name: str
        The name of this state.

    """

    def __init__(self, rgc: RunGovernor, name: str):
        self.rgc = rgc
        self.name = name

    @abstractmethod
    async def run(self) -> None:
        """The main function of this state.

        All commands of this state should be executed here. The last
        command should be :meth:`next_state` to initialize the
        transition to the following state.

        """
        pass

    @abstractmethod
    def next_state(self) -> None:
        """Transition to the next state.

        Based on the outcome of the :meth:`run` method, the next state
        should be determined in this function.
        """
        pass

    def add_error(self, error: Exception) -> None:
        """Add an error to the error list of the run governor.

        Parameters
        ----------
        error: Exception
            The error that raised, catched, and should be saved for
            error handling.

        """
        self.rgc.errors.append((error, self.name))



from dataclasses import dataclass
from typing import Any


@dataclass
class SimulationControllerTerminationRequest:
    """Announces that a :py:class:`SimulationController` has exited

    * Sender: :py:class:`SimulationController`
    * Receiver: :py:class:`RunGovernor`

    :param sender_simulation_controller_id: Opaque ID of the sending
        :py:class:`SimulationController` instance
    :param receiver_run_governor_id: Opaque ID of the receiving
        :py:class:`RunGovernor` instance
    :param experiment_run_id: Opaque ID of an experiment run
    :param environment_terminated: `True` if the environment terminated the
        run, `False` if something else (e.g., agent reset) issued the
        termination of the simulation run.
    :param additional_results: Any additional results the simulation
        communicates to the :py:class:`RunGovernor`, e.g., for a
        :py:class:`TerminationCondition`
    :param last_reward: Last reward issued in the simulation
    """

    sender_simulation_controller_id: str
    receiver_run_governor_id: str
    experiment_run_id: str
    environment_terminated: bool
    additional_results: Any
    last_reward: Any

    @property
    def sender(self):
        return self.sender_simulation_controller_id

    @property
    def receiver(self):
        return self.receiver_run_governor_id



from abc import ABC, abstractmethod


class TerminationCondition(ABC):
    """
    The base class for termination conditions.

    Parameters
    ----------

    message:
        termination request message object

    component:
        Component for which the termination condition is to be checked. Only
        needed if terminiation condition can not be derived from termination
        request alone.


    """

    @abstractmethod
    def check_termination(self, message, component):
        raise NotImplementedError



import re
import numpy as np
from .space import Space


class Discrete(Space):
    """A discrete space in :math:`\{ 0, 1, \dots, n-1 \}`.

    Example::

        >>> Discrete(2)

    """

    _RE = re.compile(r"\A\s*?Discrete\((\d+)\)\s*\Z")

    def __init__(self, n):
        assert n >= 0
        self.n = n
        super(Discrete, self).__init__((), np.int64)

    def sample(self):
        return self.np_random.randint(self.n)

    def contains(self, x):
        if isinstance(x, int):
            as_int = x
        elif isinstance(x, (np.generic, np.ndarray)) and (
            x.dtype.kind in np.typecodes["AllInteger"] and x.shape == ()
        ):
            as_int = int(x)
        else:
            return False
        return as_int >= 0 and as_int < self.n

    def to_string(self):
        return self.__repr__()

    def __repr__(self):
        return f"Discrete({self.n})"

    @classmethod
    def from_string(cls, s):
        match = Discrete._RE.match(s)
        if not match or not match[1]:
            raise RuntimeError(
                "String '%s' did not match '%s'" % (s, Discrete._RE)
            )
        return Discrete(int(match[1]))

    def __eq__(self, other):
        return isinstance(other, Discrete) and self.n == other.n



from . import LOG


class AgentShutdownResponse:
    def __init__(self, run_id, agent_id, termination):
        self._experiment_run_id = run_id
        self._agent_id = agent_id
        self._termination = termination

    @property
    def run_id(self):
        """Deprecated: Use experiment_run_id instead."""
        LOG.debug(
            f"Run_id property deprecated in class {self.__class__}. Use experiment_run_id instead."
        )
        return self.experiment_run_id

    @run_id.setter
    def run_id(self, value):
        """Deprecated: Use experiment_run_id instead."""
        LOG.debug(
            f"Run_id property deprecated in class {self.__class__}. Use experiment_run_id instead."
        )
        self.experiment_run_id(value)

    @property
    def experiment_run_id(self):
        return self._experiment_run_id

    @experiment_run_id.setter
    def experiment_run_id(self, value):
        self._experiment_run_id = value

    @property
    def agent_id(self):
        return self._agent_id

    @agent_id.setter
    def agent_id(self, value):
        self._agent_id = value

    @property
    def termination(self):
        return self._termination

    @termination.setter
    def termination(self, value):
        self._termination = value



import unittest
from asyncio import Future
from unittest.mock import AsyncMock, MagicMock, PropertyMock, patch

from palaestrai.core.protocol import (
    ExperimentRunShutdownRequest,
    ExperimentRunStartRequest,
    SimulationControllerTerminationRequest,
)
from palaestrai.experiment import RunGovernor
from palaestrai.experiment.rungovernor_states import (
    RGSErrorHandlingTransceiving,
    RGSHandlingSimControllerTermination,
    RGSStartingSetup,
    RGSStoppingSimulation,
    RGSTransceiving,
)
from palaestrai.util.exception import (
    DeadChildrenRisingAsZombiesError,
    InvalidRequestError,
    RequestIsNoneError,
    SignalInterruptError,
    TasksNotFinishedError,
)


class TestTransceiving(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.rungov = RunGovernor(None, None)
        self.rgs = RGSTransceiving(self.rungov)
        self.rungov.state = self.rgs

    async def test_run_success(self):
        self.rgs._create_transceive_task = MagicMock()
        self.rgs._wait_for_tasks = AsyncMock(
            return_value=(list(), list(), list())
        )
        self.rgs._tasks_are_done = MagicMock(return_value=True)
        self.rgs._has_dead_children = MagicMock(return_value=False)
        self.rgs._get_request = AsyncMock()

        await self.rgs.run()

        self.rgs._create_transceive_task.assert_called_once()
        self.rgs._wait_for_tasks.assert_called_once()
        self.rgs._tasks_are_done.assert_called_once()
        self.rgs._has_dead_children.assert_called_once()
        self.rgs._get_request.assert_called_once()

    async def test_run_tasks_not_done(self):
        self.rgs._create_transceive_task = MagicMock()
        self.rgs._wait_for_tasks = AsyncMock(
            return_value=(list(), list(), list())
        )
        self.rgs._tasks_are_done = MagicMock(return_value=False)
        self.rgs._has_dead_children = MagicMock(return_value=False)
        self.rgs._get_request = AsyncMock()

        await self.rgs.run()

        self.rgs._create_transceive_task.assert_called_once()
        self.rgs._wait_for_tasks.assert_called_once()
        self.rgs._tasks_are_done.assert_called_once()
        self.rgs._has_dead_children.assert_not_called()
        self.rgs._get_request.assert_not_called()

    async def test_run_childrens_dead(self):
        self.rgs._create_transceive_task = MagicMock()
        self.rgs._wait_for_tasks = AsyncMock(
            return_value=(list(), list(), list())
        )
        self.rgs._tasks_are_done = MagicMock(return_value=True)
        self.rgs._has_dead_children = MagicMock(return_value=True)
        self.rgs._get_request = AsyncMock()

        await self.rgs.run()

        self.rgs._create_transceive_task.assert_called_once()
        self.rgs._wait_for_tasks.assert_called_once()
        self.rgs._tasks_are_done.assert_called_once()
        self.rgs._has_dead_children.assert_called_once()
        self.rgs._get_request.assert_not_called()

    def test_next_state_experiment_start(self):

        self.rungov.last_request.append(
            ExperimentRunStartRequest(
                sender_executor_id="0",
                receiver_run_governor_id="1",
                experiment_run_id="2",
                experiment_run=None,
            )
        )

        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSStartingSetup)

    def test_next_state_sim_controller_termination(self):

        self.rungov.last_request.append(
            SimulationControllerTerminationRequest(
                None, None, None, None, None, None
            )
        )

        self.rgs.next_state()

        self.assertIsInstance(
            self.rungov.state, RGSHandlingSimControllerTermination
        )

    def test_next_state_experiment_shutdown(self):

        self.rungov.last_request.append(
            ExperimentRunShutdownRequest(
                sender_executor_id="0",
                receiver_run_governor_id="1",
                experiment_run_id="2",
            )
        )

        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSStoppingSimulation)

    def test_next_state_invalid_request(self):

        self.rungov.last_request.append(0)

        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSErrorHandlingTransceiving)
        self.assertEqual(len(self.rungov.errors), 1)
        self.assertIsInstance(self.rungov.errors[0][0], InvalidRequestError)

    def test_next_state_error_occured(self):
        self.rungov.last_request.append(
            ExperimentRunStartRequest(
                sender_executor_id="0",
                receiver_run_governor_id="1",
                experiment_run_id="2",
                experiment_run=None,
            )
        )
        self.rgs.add_error(DeadChildrenRisingAsZombiesError())

        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSErrorHandlingTransceiving)
        self.assertEqual(len(self.rungov.errors), 1)
        self.assertIsInstance(
            self.rungov.errors[0][0], DeadChildrenRisingAsZombiesError
        )

    @patch(f"{RGSTransceiving.__module__}.asyncio")
    def test_create_transceive_task(self, mock_asyncio):
        self.rungov.major_domo_worker = MagicMock()
        self.rungov.major_domo_worker.transceive = MagicMock()
        mock_asyncio.create_task = MagicMock()
        self.rungov.next_reply.append(None)
        self.rungov.transceive_task = None

        self.rgs._create_transceive_task()

        self.assertIsNotNone(self.rungov.transceive_task)
        self.rungov.major_domo_worker.transceive.assert_called_with(None)
        mock_asyncio.create_task.assert_called_once()

    @patch(f"{RGSTransceiving.__module__}.asyncio")
    async def test_wait_for_tasks(self, mock_asyncio):
        sim = MagicMock()
        type(sim).name = PropertyMock(return_value="sim-1")
        sim.is_alive = MagicMock(return_value=True)
        self.rungov.sim_controllers["sim-1"] = sim

        mock_asyncio.wait = AsyncMock(return_value=(list(), 0))
        self.rgs._join_process = MagicMock()

        await self.rgs._wait_for_tasks()

        sim.is_alive.assert_called_once()
        mock_asyncio.wait.assert_called_once()
        self.rgs._join_process.assert_called_once()

    def test_tasks_are_done(self):
        res = self.rgs._tasks_are_done(list("I am a task"))

        self.assertTrue(res)
        self.assertEqual(0, len(self.rungov.errors))

    def test_tasks_are_not_done(self):
        res = self.rgs._tasks_are_done(list())

        self.assertFalse(res)
        self.assertEqual(1, len(self.rungov.errors))
        self.assertIsInstance(self.rungov.errors[0][0], TasksNotFinishedError)

    def test_signal_task_is_done(self):
        self.rungov.signal_monitor_task = "signal"
        res = self.rgs._tasks_are_done(["signal"])

        self.assertFalse(res)
        self.assertEqual(1, len(self.rungov.errors))
        self.assertIsInstance(self.rungov.errors[0][0], SignalInterruptError)

    def test_childrens_are_not_dead(self):
        sim = MagicMock()
        sim.is_alive = MagicMock(return_value=True)

        res = self.rgs._has_dead_children([sim])

        self.assertFalse(res)
        self.assertEqual(0, len(self.rungov.errors))

    def test_childrens_are_dead(self):
        sim = MagicMock()
        sim.is_alive = MagicMock(return_value=False)

        res = self.rgs._has_dead_children([sim])

        self.assertTrue(res)
        self.assertEqual(1, len(self.rungov.errors))
        self.assertIsInstance(
            self.rungov.errors[0][0], DeadChildrenRisingAsZombiesError
        )

    async def test_get_request(self):

        self.rungov.transceive_task = MagicMock(return_value=Future())
        self.rungov.transceive_task.result = MagicMock()

        res = await self.rgs._get_request([self.rungov.transceive_task])

        self.assertTrue(res)

        self.assertEqual(0, len(self.rungov.errors))
        self.assertIsNone(self.rungov.transceive_task)
        self.assertEqual(1, len(self.rungov.last_request))

    async def test_get_request_is_none(self):

        self.rungov.transceive_task = MagicMock(return_value=Future())
        self.rungov.transceive_task.result = MagicMock(return_value=None)

        res = await self.rgs._get_request([self.rungov.transceive_task])

        self.assertFalse(res)

        self.assertEqual(1, len(self.rungov.errors))
        self.assertIsInstance(self.rungov.errors[0][0], RequestIsNoneError)
        self.assertIsNone(self.rungov.transceive_task)
        self.assertEqual(1, len(self.rungov.last_request))

    @patch(f"{RGSTransceiving.__module__}.asyncio")
    async def test_join_process(self, mock_asyncio):
        proc = MagicMock()
        proc.name.return_value = "proc-1"
        proc.is_alive = MagicMock(return_value=True)
        type(proc).exitcode = PropertyMock(side_effect=[None, 1, 1])
        mock_asyncio.sleep = AsyncMock()

        await self.rgs._join_process(proc)

        mock_asyncio.sleep.assert_called_once()
        self.assertEqual(2, proc.is_alive.call_count)


if __name__ == "__main__":
    unittest.main()



import unittest

from palaestrai.experiment import RunGovernor
from palaestrai.experiment.rungovernor_states import RGSDone


class TestDone(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.rungov = RunGovernor(None, None)
        self.rgs = RGSDone(self.rungov)
        self.rungov.state = self.rgs

    async def test_run(self):

        await self.rgs.run()

        self.assertTrue(self.rungov.shutdown)
        self.assertEqual(0, len(self.rungov.errors))

    def test_next_state(self):

        # Nothing to test here
        self.rgs.next_state()


if __name__ == "__main__":
    unittest.main()



from palaestrai.agent import RewardInformation
from palaestrai.agent.objective import Objective
from typing import List


class DummyObjective(Objective):
    def __init__(self, params):
        super().__init__(params)

    def internal_reward(self, rewards: List[RewardInformation]):
        final_rewards = []
        for reward in rewards:
            r = reward.reward_value
            final_rewards.append(r)
        return sum(final_rewards)



from typing import List

from palaestrai.agent import SensorInformation, RewardInformation
from palaestrai.environment.reward import Reward
from palaestrai.types import Discrete


class DummyReward(Reward):
    def __call__(
        self, state: List[SensorInformation], *args, **kwargs
    ) -> List[RewardInformation]:
        reward = 0
        for sensor in state:
            reward += sensor()
        if reward < 9000:
            return [RewardInformation(reward, Discrete(9000), "Reward")]
        else:
            # This reward is over 9000, maybe add a log
            return [RewardInformation(9000, Discrete(9000), "Reward")]



from __future__ import annotations

import io
import signal
import traceback
from typing import TYPE_CHECKING

import aiomultiprocess
import setproctitle

from ...core import RuntimeConfig
from ...util import spawn_wrapper
from . import LOG, RunGovernorState

if TYPE_CHECKING:
    from ...agent import AgentConductor
    from ...experiment import RunGovernor


async def _run_agent_conductor(agent_conductor: AgentConductor):
    """Executes the :py:class:`AgentConductor` main loop

    This is a wrapper function around :py:func:`AgentConductor.run`.
    It takes care of clearing signal handlers, setting the proctitle,
    and generally catching errors in a meaningful in order to report it
    to the RunGovernor without simply dying.

    Parameters
    ----------
    agent_conductor : ::`~AgentConductor`
        An initialized agent conductor, ready to be run.

    Returns
    -------
    Any
        Whatever the :py:func:`AgentConductor.run` method returns
    """
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    signal.signal(signal.SIGABRT, signal.SIG_DFL)
    signal.signal(signal.SIGTERM, signal.SIG_DFL)
    setproctitle.setproctitle(
        "palaestrAI[AgentConductor-%s]" % agent_conductor.uid[-6:]
    )
    try:
        rc = await agent_conductor.run()
        LOG.debug(
            "AgentConductor(id=0x%x, uid=%s) exited normally.",
            id(agent_conductor),
            agent_conductor.uid,
        )
        return rc
    except Exception as e:
        LOG.critical(
            "AgentConductor(id=0x%x, uid=%s) died orchestrating: %s",
            id(agent_conductor),
            agent_conductor.uid,
            e,
        )
        tb = io.StringIO()
        traceback.print_exc(file=tb)
        LOG.debug(
            "AgentConductor(id=0x%x, uid=%s) died orchestrating: %s",
            id(agent_conductor),
            agent_conductor.uid,
            tb.getvalue(),
        )
        raise


class RGSStartingAgentConductors(RunGovernorState):
    """Represent the STARTING_AGENT_CONDUCTORS state of the run
    governor.

    This state is entered after STARTING_ENV_CONDUCTORS succeeds.
    During this state, processes for all required agent conductors are
    created (but not started).

    Possible next states are
    * :class:`.RGSStartingRun in the normal case
    * :class:`.RGSErrorHandlingStarting in the error case

    Parameters
    ----------
    rgc: :class:`.RunGovernor`
        The run governor instance that provides the context for this
        state.


    """

    def __init__(self, rgc: "RunGovernor"):
        super().__init__(rgc, "STARTING_AGENT_CONDUCTORS")

    async def run(self):
        self._init_agent_conductors()

    def next_state(self, *args, **kwargs):
        from . import RGSErrorHandlingStarting, RGSStartingRun

        if len(self.rgc.errors) > 0:
            self.rgc.state = RGSErrorHandlingStarting(self.rgc)
        else:
            self.rgc.state = RGSStartingRun(self.rgc)

    def _init_agent_conductors(self):
        """Initializes and starts given agent conductors as subprocesses.

        All objects of type :py:class:`AgentConductor` are treated as
        subprocesses. As long as they run on the same machine, they are
        initialized in this method.

        The processes are not started.

        This method accesses ::`~RunGovernor.experiment.agent_conductors` to
        obtain all ::`AgentConductor` objects. It sets
        ::`~RunGovernor.agent_conductors`.
        """
        agent_conductors = list(
            self.rgc.experiment_run.agent_conductors(
                self.rgc.current_phase
            ).values()
        )
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) initializing %d agent conductor(s).",
            id(self.rgc),
            self.rgc.uid,
            len(agent_conductors),
        )
        self.rgc.agent_conductors = {
            ac.uid: aiomultiprocess.Process(
                name=f"AgentConductor-{ac.uid}",
                target=spawn_wrapper,
                args=(
                    f"AgentConductor-{ac.uid}",
                    RuntimeConfig().to_dict(),
                    _run_agent_conductor,
                    [ac],
                ),
            )
            for ac in agent_conductors
        }



class AgentShutdownRequest:
    def __init__(self, run_id, agent_id, complete_shutdown):
        self.experiment_run_id = run_id
        self.agent_id = agent_id
        self.complete_shutdown = complete_shutdown



from itertools import product

from palaestrai.core.protocol import (
    AgentUpdateRequest,
    AgentUpdateResponse,
    EnvironmentResetNotificationRequest,
    EnvironmentResetNotificationResponse,
    EnvironmentResetRequest,
    EnvironmentResetResponse,
    EnvironmentUpdateRequest,
    EnvironmentUpdateResponse,
    SimulationControllerTerminationRequest,
    SimulationControllerTerminationResponse,
)
from . import LOG
from .simulation_controller import SimulationController


class VanillaSimController(SimulationController):
    """
    This is our vanilla controller. With this simulation,
    environment and agent(s) work alternately and not in
    parallel. It is not a continuous simulation.
    After each simulation step (independent if done
    by environment or agent) the Termination Condition
    is called.
    """

    def __init__(
        self,
        rungov_connection,
        sim_connection,
        agent_conductor_ids,
        environment_conductor_ids,
        agents,
        termination_conditions,
        mode,
    ):
        super().__init__(
            rungov_connection,
            sim_connection,
            agent_conductor_ids,
            environment_conductor_ids,
            agents,
            termination_conditions,
            mode,
        )

    @staticmethod
    def env_update(responses: list):
        """
        This method will process a list of environment update responses
        and will combine the information to combined lists. Currently
        the termination variable is global because we assume, that
        the run terminates if at least one environment is env
        terminates.

        Parameters
        ----------
        responses = list of EnvironmentUpdateResponse

        Returns
        -------
        sensors = list of all sensor_information of all available environments
        rewards = list of all environment rewards
        termination = boolean which is true if one environment has terminated

        """
        sensors = []
        rewards = []
        termination = False
        for response in responses:
            sensors.extend(response.sensors)
            rewards.extend(response.reward)
            if response.is_terminal:
                termination = True
        return sensors, rewards, termination

    @staticmethod
    def agent_update(responses: list) -> list:
        """

        This method combines all actuator_information of all Agents
        and creates on list.

        Parameters
        ----------
        responses = List of AgentUpdateResponses

        Returns
        -------
        actuators which is a list of actuator_information
        """
        actuators = []
        for response in responses:
            actuators.extend(response.actuators)
        return actuators

    async def simulation_shutdown(
        self, env_termination: bool, rewards: list, additional_results
    ):
        """
        This method will be called when the simulation has terminated
        it will send a SimControllerTerminationRequest to the runGov.
        The RunGov will respond and will tell if it will be a complete
        or partial shutdown.
        The complete shutdown includes the conductors while a partial
        shutdown is a reset which just deletes the muscle(s) and env(s)
        Parameters
        ----------
        env_termination : bool if the environment has terminated
        rewards : list of rewards to show the current performance
        additional_results : for any additional information

        Returns
        -------


        """

        msg = SimulationControllerTerminationRequest(
            sender_simulation_controller_id=self.uid,
            receiver_run_governor_id=self.rg_id,
            experiment_run_id=self.experiment_run_id,
            environment_terminated=env_termination,
            last_reward=rewards,
            additional_results=additional_results,
        )
        LOG.debug(
            "SimulationController %s(id=0x%x, uid=%s) sending "
            "SimulationControllerTerminationRequest(experiment_run_id=%s).",
            self.__class__,
            id(self),
            self.uid,
            self.experiment_run_id,
        )

        response = await self.run_gov_client.send(self.rg_id, msg)
        if not isinstance(response, SimulationControllerTerminationResponse):
            LOG.critical(
                "SimulationController %s(id=0x%x, uid=%s) "
                "waited for SimulationControllerTerminationResponse, but got "
                "%s instead. Dying without honor, trusting the RunGovernor "
                "to handle this disgrace.",
                self.__class__,
                id(self),
                self.uid,
                response,
            )
            await self.stop_simulation(True)
        LOG.debug(
            "SimulationController %s(id=0x%x, uid=%s) received %s.",
            self.__class__,
            id(self),
            self.uid,
            response,
        )
        if response.complete_shutdown:
            await self.stop_simulation(response.complete_shutdown)
        # TODO: if episodes > 1: restart
        else:
            await self.agent_shutdown(response.complete_shutdown)
            await self.env_shutdown()

    async def get_env_update(self, env, actuators):
        """
        Sends an EnvironmentUpdateRequest to one env
        and collects the Response.
        The vanilla simController sends all actuators to all envs
        and the env has to select the own actuators. A access list
        could be needed if two envs of the same type are used.
        Parameters
        ----------
        env : id of the environment
        actuators : list of actuatorinformation

        Returns
        -------
        response : EnvironmentUpdateResponse
        """
        LOG.debug(
            "SimulationController %s(id=0x%x, uid=%s) "
            "starting EnvironmentUpdateRequest(experiment_run_id=%s, "
            "env=%s, actuators=%s).",
            self.__class__,
            id(self),
            self.uid,
            self.experiment_run_id,
            str(env),
            actuators,
        )
        msg = EnvironmentUpdateRequest(
            experiment_run_id=self.experiment_run_id,
            environment_conductor_id=str(self.env_ids[str(env)]),
            sender_simulation_controller=self.uid,
            receiver_environment=str(env),
            actuators=actuators,
        )
        response = await self.client.send(bytes(str(env), "ascii"), msg)
        LOG.debug(
            "SimulationController %s(id=0x%x, uid=%s) "
            "received EnvironmentUpdateResponse: %s.",
            self.__class__,
            id(self),
            self.uid,
            response,
        )
        if isinstance(response, EnvironmentUpdateResponse):
            return response
        else:
            LOG.error(
                "SimulationController %s(id=0x%x, uid=%s) expected "
                "EnvironmentUpdateResponse, but got %s instead; ignoring.",
                self.__class__,
                id(self),
                self.uid,
                response,
            )
            return None

    async def get_agent_update(
        self, agent_id, sensors, actuators, rewards, env_termination
    ):
        """
        This method sends an AgentUpdateRequest and collects
        the response which will be returned.

        Parameters
        ----------
        agent_id : id of the agent
        sensors : list of sensor information
        actuators : list of actuator information
        rewards : list of environment rewards
        env_termination : bool if env has terminated

        Returns
        -------
        AgentUpdateResponse
        """
        LOG.debug(
            "SimulationController %s(id=0x%x, uid=%s) "
            "requesting update from Agent(uid=%s).",
            self.__class__,
            id(self),
            self.uid,
            agent_id,
        )
        msg = AgentUpdateRequest(
            sender_simulation_controller_id=self.uid,
            receiver_agent_id=agent_id,
            experiment_run_id=self.experiment_run_id,
            sensors=sensors,
            actuators=actuators,
            reward=rewards,
            is_terminal=env_termination,
            mode=self.mode,
        )
        response = await self.client.send(agent_id, msg)

        if isinstance(response, AgentUpdateResponse):
            LOG.debug(
                "SimulationController %s(id=0x%x, uid=%s) "
                "received AgentUpdateResponse from Agent(uid=%s).",
                self.__class__,
                id(self),
                self.uid,
                agent_id,
            )
            # TODO: this is quite hacky
            if len(response.actuators) > 0:
                for agent in self.agents:
                    if agent.uid == agent_id:
                        agent.actuators = response.actuators
                        break
            return response
        else:
            LOG.error(
                "SimulationController %s(id=0x%x, uid=%s) expected "
                "AgentUpdateResponse, but got %s instead; ignoring.",
                self.__class__,
                id(self),
                self.uid,
                response,
            )
            return None

    async def simulation(self):
        """
        This is the abstract method implementation of the
        simulation task. The vanilla sim controller simulation
        start by asking all environments for sensor information.
        These information will be send to the agent(s) which
        will respond which their actuator-setpoints.
        From there on it will be a ping pong between environment(s)
        and agent(s). All available information will be exchanged.
        Both, environment as well as agent information, can be
        used for the termination.
        Returns
        -------

        """
        termination = False
        actuators = self.all_actuators
        env_termination = False
        rewards = []
        LOG.info(
            "SimulationController %s(id=0x%x, uid=%s) starting simulation.",
            self.__class__,
            id(self),
            self.uid,
        )
        while True:
            LOG.debug(
                "SimulationController %s(id=0x%x, uid=%s) "
                "running new iteration for experiment_run_id=%s; "
                "termination: %s.",
                self.__class__,
                id(self),
                self.uid,
                self.experiment_run_id,
                self.ex_termination,
            )

            responses = [
                term
                for term in [
                    await self.get_env_update(eid, actuators)
                    for eid in self.env_ids
                ]
                if term
            ]
            termination = any(
                [
                    term[0].check_termination(term[1])
                    for term in product(self.termination_conditions, responses)
                ]
            )
            self.all_sensors, rewards, env_termination = self.env_update(
                responses
            )

            responses = [
                term
                for term in [
                    await self.get_agent_update(
                        agent_id=str(agent.uid),
                        sensors=self.access_list(
                            [s.sensor_id for s in agent.sensors], []
                        )[
                            0
                        ],  # TODO: quite hacky
                        actuators=agent.actuators,
                        rewards=rewards,
                        env_termination=env_termination,
                    )
                    for agent in self.agents
                ]
                if term
            ]
            termination = termination or any(
                [
                    term[0].check_termination(term[1])
                    for term in product(self.termination_conditions, responses)
                ]
            )
            actuators = self.collect_actuators(agent_responses=responses)

            if not await self.continue_simulation(termination, rewards, None):
                break

        LOG.info(
            "SimulationController %s(id=0x%x, uid=%s) " "finished simulation.",
            self.__class__,
            id(self),
            self.uid,
        )

    async def continue_simulation(
        self, termination, rewards, additional_results
    ):
        if self.ex_termination:
            LOG.info(
                "SimulationController %s(id=0x%x, uid=%s) encountered "
                "external termination. Performing complete shutdown!",
                self.__class__,
                id(self),
                self.uid,
            )
            await self.stop_simulation(True)
            return False
        elif termination:
            LOG.info(
                "SimulationController %s(id=0x%x, uid=%s) encountered "
                "internal termination.",
                self.__class__,
                id(self),
                self.uid,
            )
            response = await self.send_termination_request(
                rewards, additional_results
            )
            if response.complete_shutdown:
                await self.stop_simulation(True)
                return False
            elif response.restart:
                LOG.info(
                    "SimulationController %s/id=0x%x, uid=%s) initiating "
                    "restart.",
                    self.__class__,
                    id(self),
                    self.uid,
                )
                await self.perform_restart()
                return True
            else:
                await self.agent_shutdown(response.complete_shutdown)
                await self.env_shutdown()
                # TODO: This situation is not handled in the main loop
                # Therefore, we will just break the simulation loop.
                return False
        else:
            return True

    async def send_termination_request(
        self, rewards: list, additional_results
    ):
        msg = SimulationControllerTerminationRequest(
            sender_simulation_controller_id=self.uid,
            receiver_run_governor_id=self.rg_id,
            experiment_run_id=self.experiment_run_id,
            environment_terminated=True,
            last_reward=rewards,
            additional_results=additional_results,
        )
        LOG.debug(
            "SimulationController %s(id=0x%x, uid=%s) sending "
            "SimulationControllerTerminationRequest(experiment_run_id=%s).",
            self.__class__,
            id(self),
            self.uid,
            self.experiment_run_id,
        )
        response = await self.run_gov_client.send(self.rg_id, msg)
        if not isinstance(response, SimulationControllerTerminationResponse):
            LOG.critical(
                "SimulationController %s(id=0x%x, uid=%s) "
                "waited for SimulationControllerTerminationResponse, but got "
                "%s instead. Dying without honor, trusting the RunGovernor "
                "to handle this disgrace.",
                self.__class__,
                id(self),
                self.uid,
                response,
            )
            await self.stop_simulation(True)
        LOG.debug(
            "SimulationController %s(id=0x%x, uid=%s) received %s.",
            self.__class__,
            id(self),
            self.uid,
            response,
        )
        return response

    async def perform_restart(self):
        for env_id in self.env_ids:
            msg = EnvironmentResetRequest(
                sender_simulation_controller_id=self.uid,
                receiver_environment_id=env_id,
            )
            LOG.debug(
                "SimulationController %s(id=0x%x, uid=%s) sending "
                "EnvironmentResetRequest(env.uid=%s).",
                self.__class__,
                id(self),
                self.uid,
                env_id,
            )
            response = await self.client.send(env_id, msg)
            if isinstance(response, EnvironmentResetResponse):
                if response.create_new_instance:
                    LOG.warning(
                        "Environment (%s) requested a new instance. "
                        "However, this feature is not implemented, yet.",
                        response.sender,
                    )
                    # TODO: notify conductor to reinstantiate env
            else:
                # TODO: notify conductor to reinstantiate env
                pass

        for agent in self.agents:
            # TODO: We could send the new sensors and actuators
            # but normally a reset should not change them.
            msg = EnvironmentResetNotificationRequest(
                sender_simulation_controller_id=self.uid,
                receiver_agent_id=agent.uid,
            )
            LOG.debug(
                "SimulationController %s(id=0x%x, uid=%s) sending "
                "EnvironmentResetNotificationRequest(agent.uid=%s).",
                self.__class__,
                id(self),
                self.uid,
                agent.uid,
            )
            response = await self.client.send(agent.uid, msg)
            if isinstance(response, EnvironmentResetNotificationResponse):
                # TODO: We could allow the agent do provide additional info
                LOG.debug(
                    "SimulationController(id=0x%x, uid=%s) received "
                    "EnvironmentResetNotificationResponse "
                    "but no procedure is implemented to handle it.",
                    id(self),
                    self.uid,
                )
            else:
                LOG.warning(
                    "SimulationController(id=0x%x, uid=%s) expected "
                    "EnvironmentResetNotificationResponse but got %s instead.",
                    id(self),
                    self.uid,
                    response,
                )

    def collect_actuators(self, agent_responses: list):
        """
        collect_actuators takes a list of agent_responses
        and combines all available actuators to one list.

        Parameters
        ----------
        agent_responses : list[agent_responses]

        Returns
        -------
        actuators : list[actuator]

        """
        actuators = []
        for response in agent_responses:
            # TODO: Workaround, works only for single env
            actuators.extend(response.actuators)
        return actuators

    def access_list(self, sensor_list, actuator_list):
        """
        access_list takes a list of sensors and actuators
        and checks if they are part of the available
        sensors/actuators. If that is the case they
        will be returned.

        Parameters
        ----------
        sensor_list : list
            a list with sensor IDs.
        actuator_list : list
            a list with actuator IDs.
        Returns
        -------
        sensor_list : list
        actuator_list : list

        """
        return (
            [s for s in self.all_sensors if s.sensor_id in sensor_list],
            [a for a in self.all_actuators if a.actuator_id in actuator_list],
        )



class EnvironmentStartResponse:
    def __init__(self, run_id, environment_id, sensors, actuators):
        self._run_id = run_id
        self._environment_id = environment_id
        self._sensors = sensors
        self._actuators = actuators

    @property
    def run_id(self):
        return self._run_id

    @run_id.setter
    def run_id(self, value):
        self._run_id = value

    @property
    def environment_id(self):
        return self._environment_id

    @environment_id.setter
    def environment_id(self, value):
        self._environment_id = value

    @property
    def actuators(self):
        return self._actuators

    @actuators.setter
    def actuators(self, value):
        self._actuators = value

    @property
    def sensors(self):
        return self._sensors

    @sensors.setter
    def sensors(self, value):
        self._sensors = value



import logging

LOG = logging.getLogger(__name__)

from .agent import Agent
from .brain import Brain
from .muscle import Muscle
from .objective import Objective
from .dummy_brain import DummyBrain
from .dummy_muscle import DummyMuscle
from .agent_conductor import AgentConductor
from .sensor_information import SensorInformation
from .actuator_information import ActuatorInformation
from .reward_information import RewardInformation



class SimulationStartRequest:
    """Requests a :class:`SimulationController` to start a simulation run.

    * Sender: :class:`RunGovernor`
    * Receiver: :class:`SimulationController`

    :param run_governor_id: ID of the :class:`RunGovernor` that requests the
        simulation start (i.e., parent of the :class:`SimulationController`)
    :param simulation_controller_id: Target :class:`SimulationController`
    :param experiment_run_id: ID of the experiment run being conducted
    """

    def __init__(
        self, run_governor_id, simulation_controller_id, experiment_run_id
    ):
        self.run_governor_id = run_governor_id
        self.experiment_run_id = experiment_run_id
        self.simulation_controller_id = simulation_controller_id



import unittest

import aiomultiprocess

from palaestrai.core import RuntimeConfig
from palaestrai.util import spawn_wrapper


def _dummy_spawn_target(x=0, y=0):
    c = RuntimeConfig()
    if c.store_uri != "bar":
        exit(5)
    exit(x + y)


async def _async_dummy_spawn_target(x=0, y=0):
    return _dummy_spawn_target(x, y)


class SpawnTest(unittest.IsolatedAsyncioTestCase):
    async def test_spawn_wrapper(self):
        RuntimeConfig()._store_uri = "bar"
        p = aiomultiprocess.Process(
            target=spawn_wrapper,
            args=(
                __name__,
                RuntimeConfig().to_dict(),
                _dummy_spawn_target,
            ),
        )
        p.start()
        await p.join()
        self.assertEqual(p.exitcode, 0)

    async def test_spawn_wrapper_no_args(self):
        RuntimeConfig()._store_uri = "bar"
        p = aiomultiprocess.Process(
            target=spawn_wrapper,
            args=(
                __name__,
                RuntimeConfig().to_dict(),
                _dummy_spawn_target,
                [5],
                {"y": 37},
            ),
        )
        p.start()
        await p.join()
        self.assertEqual(p.exitcode, 42)

    async def test_spawn_wrapper_coroutine(self):
        RuntimeConfig()._store_uri = "bar"
        p = aiomultiprocess.Process(
            target=spawn_wrapper,
            args=(
                __name__,
                RuntimeConfig().to_dict(),
                _async_dummy_spawn_target,
                [5],
                {"y": 37},
            ),
        )
        p.start()
        await p.join()
        self.assertEqual(p.exitcode, 42)


if __name__ == "__main__":
    unittest.main()



import io
import traceback
from typing import TYPE_CHECKING, Union

from palaestrai.core.protocol import (
    ExperimentRunStartRequest,
    ExperimentRunStartResponse,
    NextPhaseRequest,
    NextPhaseResponse,
)

if TYPE_CHECKING:
    from palaestrai.experiment import RunGovernor

from palaestrai.util.exception import (
    ExperimentAlreadyRunningError,
    ExperimentSetupFailedError,
    InvalidRequestError,
)

from . import LOG, RunGovernorState


class RGSStartingSetup(RunGovernorState):
    """Represent the STARTING_RUN_SETUP state of the run governor.

    This state is entered if an ExperimentStartRequest was received in
    the TRANSCEIVING state or after on run phase to initiate the next
    phase. During this state, the ExperimentStartRequest is checked
    against a possibly already running experiment. If no experiment is
    running, the setup of the experiment is executed.

    In any case, an ExperimentStartResponse is prepared and stored as
    next_reply in the run governor.

    Possible next states are
    * :class:`.RGSStartinSimControllers in the normal case.
    * :class:`.RGSErrorHandlingStarting in the error case.

    Parameters
    ----------
    rgc: :class:`.RunGovernor`
        The run governor instance that provides the context for this
        state.

    """

    def __init__(self, rgc: "RunGovernor"):
        super().__init__(rgc, "STARTING_RUN_SETUP")

    async def run(self):

        try:
            request = self.rgc.last_request[0]
        except IndexError:
            # Is raised on signal interrupt because the last request
            # is none.
            return

        if isinstance(request, ExperimentRunStartRequest):
            if self._is_experiment_run_running(request):
                return

        await self._experiment_run_start(request)

    def next_state(self):
        from . import RGSErrorHandlingStarting, RGSStartingSimControllers

        if len(self.rgc.errors) > 0:
            self.rgc.state = RGSErrorHandlingStarting(self.rgc)
        else:
            self.rgc.state = RGSStartingSimControllers(self.rgc)

    def _is_experiment_run_running(
        self, request: ExperimentRunStartRequest
    ) -> bool:
        """Check if an experiment run is already being executed.

        Adds an ExperimentAlreadyRunningError to the run governor error
        stack if an experiment run is already being executed.

        Returns
        -------
        bool
            True, if an experiment run is already being executed,
            False otherwise.

        """
        if self.rgc.experiment_run is not None:
            LOG.warning(
                "RunGovernor(id=0x%x, uid=%s) received request to start "
                "ExperimentRun(run_id=%s), but this is recorded as already "
                "running.",
                id(self.rgc),
                self.rgc.uid,
                request.experiment_run_id,
            )
            self.rgc.next_reply.append(
                ExperimentRunStartResponse(
                    sender_run_governor_id=self.rgc.uid,
                    receiver_executor_id=request.sender,
                    experiment_run_id=str(self.rgc.experiment_run_id),
                    successful=False,
                    error=ExperimentAlreadyRunningError(
                        "Experiment is already active."
                    ),
                )
            )
            self.add_error(ExperimentAlreadyRunningError())
            return True
        return False

    async def _experiment_run_start(
        self, request: Union[ExperimentRunStartRequest, NextPhaseRequest]
    ):
        """Start an experiment run.

        This function starts the experiment run by receiving an
        ExperimentRunStartRequest. Adds an ExperimentSetupFailedError
        if an error occured during setup(). If the request is of
        type NextPhaseRequest, the phase counter is increased and
        some cleanup is performed.

        Parameters
        ----------
        request: Union[ExperimentStartRequest, NextPhaseRequest]
            The instance of the start request with information about
            the experiment run to start or the information about
            the next phase.

        """
        successful = True
        msg: Union[Exception, None] = None
        if isinstance(request, ExperimentRunStartRequest):
            self.rgc.experiment_run = request.experiment_run
            self.rgc.experiment_run_id = request.experiment_run_id
            try:
                self.rgc.experiment_run.setup(self.rgc.broker_uri)
            except Exception as e:
                tb = io.StringIO()
                traceback.print_exc(file=tb)
                self.add_error(ExperimentSetupFailedError(tb.getvalue()))
                msg = e
                successful = False

            self.rgc.termination_condition = (
                self.rgc.experiment_run.run_governor_termination_condition
            )

            self.rgc.next_reply.append(
                ExperimentRunStartResponse(
                    sender_run_governor_id=self.rgc.uid,
                    receiver_executor_id=request.sender,
                    experiment_run_id=self.rgc.experiment_run_id,
                    successful=successful,
                    error=msg,
                )
            )
        elif isinstance(request, NextPhaseRequest):
            LOG.info(
                "RunGovernor(id=0x%x, uid=%s) initializing next phase %d",
                id(self.rgc),
                self.rgc.uid,
                request.next_phase,
            )
            self.rgc.current_phase = request.next_phase
            self.rgc.sim_controllers = dict()
            self.rgc.agent_conductors = dict()
            self.rgc.env_conductors = dict()
            self.rgc.current_episode = 0
            if len(self.rgc.next_reply) == 0:
                self.rgc.next_reply.append(
                    NextPhaseResponse(
                        sender_run_governor_id=self.rgc.uid,
                        receiver_run_governor_id=self.rgc.uid,
                        has_next_phase=True,
                    )
                )

        else:
            # That case shouldn't be able to occur. But you never know ...
            LOG.error(
                "RunGovernor(id=0x%x, uid=%s) got invalid request "
                "during experiment run setup: %s.",
                id(self.rgc),
                self.rgc.uid,
                request,
            )
            self.rgc.add_error(
                InvalidRequestError(
                    [ExperimentRunStartRequest, NextPhaseRequest], request
                )
            )



import unittest
from unittest.mock import AsyncMock, MagicMock, PropertyMock, patch

from palaestrai.core.protocol import (
    ExperimentRunShutdownRequest,
    ExperimentRunShutdownResponse,
    NextPhaseRequest,
    NextPhaseResponse,
    SimulationStopResponse,
)
from palaestrai.experiment import RunGovernor
from palaestrai.experiment.rungovernor_states import (
    RGSStoppingSimulation,
    RGSStoppingTransceiving,
)


class TestStoppingSimulation(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.rungov = RunGovernor(None, None)
        self.rgs = RGSStoppingSimulation(self.rungov)
        self.rungov.state = self.rgs

    @patch(f"{RGSStoppingSimulation.__module__}.asyncio.wait_for")
    async def test_run(self, mock_wait):
        # self.rgs._is_experiment_running = MagicMock(return_value=False)
        # self.rgs._experiment_start = AsyncMock()
        # request = ExperimentStartRequest(None, None)
        self.rgs._prepare_reply = MagicMock()
        self.rgs._terminate_sim_controllers = MagicMock()

        # self.rungov.last_request.append(request)

        await self.rgs.run()

        self.rgs._prepare_reply.assert_called_once()
        self.rgs._terminate_sim_controllers.assert_called_once()

    def test_next_state(self):
        self.rungov.experiment_run = MagicMock()
        self.rungov.experiment_run.has_next_phase = MagicMock(
            return_value=False
        )
        self.rgs.next_state()
        self.assertIsInstance(self.rungov.state, RGSStoppingTransceiving)

    async def test_terminate_sim_controllers(self):
        self.rungov.experiment_run = MagicMock()
        type(self.rungov.experiment_run).user_id = PropertyMock(
            return_value="user_id"
        )
        self.rungov.major_domo_client = MagicMock()
        self.rungov.major_domo_client.send = AsyncMock(
            return_value=SimulationStopResponse("run-1")
        )
        proc = MagicMock()
        proc.terminate = MagicMock()

        self.rungov.sim_controllers["sim-1"] = proc
        sc_terminated = await self.rgs._terminate_sim_controllers()

        proc.terminate.assert_not_called()
        self.assertEqual(0, len(sc_terminated))

    async def test_terminate_sim_controllers_no_response(self):
        self.rungov.experiment_run = MagicMock()
        type(self.rungov.experiment_run).user_id = PropertyMock(
            return_value="user_id"
        )

        self.rungov.major_domo_client = MagicMock()
        self.rungov.major_domo_client.send = AsyncMock(return_value=None)
        proc = MagicMock()
        proc.terminate = MagicMock()

        self.rungov.sim_controllers["sim-1"] = proc
        sc_terminated = await self.rgs._terminate_sim_controllers()

        proc.terminate.assert_called_once()
        self.assertEqual(1, len(sc_terminated))

    def test_prepare_reply_shutdown(self):
        self.assertEqual(0, len(self.rungov.next_reply))
        self.rungov.last_request.append(
            ExperimentRunShutdownRequest(
                sender_executor_id="0",
                receiver_run_governor_id="1",
                experiment_run_id="2",
            )
        )
        self.rgs._prepare_reply(list())

        self.assertEqual(1, len(self.rungov.next_reply))
        self.assertIsInstance(
            self.rungov.next_reply[0], ExperimentRunShutdownResponse
        )

    def test_prepare_reply_next_phase(self):
        self.assertEqual(0, len(self.rungov.next_reply))
        self.rungov.last_request.append(
            NextPhaseRequest(
                sender_run_governor_id="1",
                receiver_run_governor_id="1",
                next_phase=1,
            )
        )
        self.rgs._prepare_reply(list())

        self.assertEqual(1, len(self.rungov.next_reply))
        self.assertIsInstance(self.rungov.next_reply[0], NextPhaseResponse)


if __name__ == "__main__":
    unittest.main()



from dataclasses import dataclass


class AgentConductorSetupRequest:
    """Initiates setup of an :class:`AgentConductor`.

    * Sender:
    * Receiver:
    """

    def __init__(self, run_id, conductor_id, agent_config):
        self._run_id = run_id
        self._agent_conductor_id = conductor_id
        self._agent_config = agent_config

    @property
    def run_id(self):
        return self._run_id

    @property
    def agent_conductor_id(self):
        return self._agent_conductor_id

    @property
    def agent_config(self):
        return self._agent_config



from .muscle import Muscle


class DummyMuscle(Muscle):
    def __init__(self, broker_connection, brain_connection, uid, brain_id):
        super().__init__(broker_connection, brain_connection, uid, brain_id)

    def setup(self):
        pass

    def propose_actions(self, sensors, actuators_available, is_terminal=False):
        for actuator in actuators_available:
            actuator(actuator.action_space.sample())
        return actuators_available, actuators_available

    def update(self, update):
        pass

    @property
    def parameters(self) -> dict:
        params = {
            "Broker_Connection": self.brain_connection,
            "Brain_Connection": self.brain_connection,
            "UID": self.uid,
        }
        return params

    def __repr__(self):
        pass

    def prepare_model(self):
        pass



from typing import TYPE_CHECKING

from . import LOG, RunGovernorState

if TYPE_CHECKING:
    from palaestrai.experiment import RunGovernor


class RGSErrorHandlingInitializing(RunGovernorState):
    """Represent the ERROR_HANDLING_INITIALIZING state of the run
    governor.

    This state handles errors that occur during initialization.
    Currently, no errors are expected in the initialization process,
    which makes this class kind of useless. But since this may change,
    we keep the state in existence.

    Possible next states are
    * :class:`.RGSStoppingTransceiving` in the normal case.
    """

    def __init__(self, rgc: "RunGovernor"):
        super().__init__(rgc, "ERROR_HANDLING_INITIALIZING")

    async def run(self):
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) encountered errors during "
            "initializing. Handling these errors now.",
            id(self.rgc),
            self.rgc.uid,
        )

    def next_state(self):
        from . import RGSStoppingTransceiving

        self.rgc.state = RGSStoppingTransceiving(self.rgc)
        self.rgc.errors = list()



import os
import sys

from palaestrai.cli.manager import cli

if __name__ == "__main__":
    sys.path.insert(
        0, os.path.abspath(os.path.join(__file__, "..", "..", ".."))
    )
    from tests.system.test_cli import debug_script_path, runtime_path

    cli(["-c", runtime_path, "experiment-start", debug_script_path])



from dataclasses import dataclass


@dataclass
class AgentSetupResponse:
    """Response to a successful agent setup

    * Sender: :class:`AgentConductor`
    * Receiver: :class:`SimulationController`

    :param sender_agent_conductor: ID of the transmitting
        :class:`AgentConductor`
    :param receiver_simulation_controller: ID of the receiving
        :class:`SimulationController`
    :param experiment_run_id: ID of the experiment run this agent will
        participate in
    :param agent_id: ID of the respective :class:`Agent` we've just set up
    """

    sender_agent_conductor: str
    receiver_simulation_controller: str
    experiment_run_id: str
    agent_id: str

    @property
    def sender(self):
        return self.sender_agent_conductor

    @property
    def receiver(self):
        return self.receiver_simulation_controller



import unittest
from unittest.mock import AsyncMock

from palaestrai.experiment import RunGovernor
from palaestrai.experiment.rungovernor_states import (
    RGSStoppingTransceiving,
    RGSStoppingRun,
)


class TestStoppingTransceiving(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.rungov = RunGovernor(None, None)
        self.rgs = RGSStoppingTransceiving(self.rungov)
        self.rungov.state = self.rgs

    async def test_run(self):
        self.rungov.major_domo_worker = AsyncMock()
        self.rungov.major_domo_worker.transceive = AsyncMock()

        await self.rgs.run()

        self.rungov.major_domo_worker.transceive.assert_called_once_with(
            None, skip_recv=True
        )

    def test_next_state(self):
        self.rgs.next_state()
        self.assertIsInstance(self.rungov.state, RGSStoppingRun)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import io
import signal
import traceback
from typing import TYPE_CHECKING

import aiomultiprocess
import setproctitle

from ...core import RuntimeConfig
from ...util.spawn import spawn_wrapper
from . import LOG, RunGovernorState

if TYPE_CHECKING:
    from ...experiment import RunGovernor
    from ...simulation import SimulationController


async def _run_simulation_controller(
    simulation_controller: SimulationController,
):
    """Executes the :py:class:`SimulationController` main loop

    This is a wrapper function around :py:func:`SimulationController.run`.
    It takes care of clearing signal handlers, setting the proctitle,
    and generally catching errors in a meaningful in order to report it
    to the RunGovernor without simply dying.

    Parameters
    ----------
    simulation_controller : palaestrai.simulation.SimulationController
        An initialized simulation controller, ready to be run.

    Returns
    -------
    Any
        Whatever the concrete :py:func:`SimulationController.run`
        method returns.
    """
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    signal.signal(signal.SIGABRT, signal.SIG_DFL)
    signal.signal(signal.SIGTERM, signal.SIG_DFL)
    setproctitle.setproctitle(
        "palaestrAI[SimulationController-%s]" % simulation_controller.uid[-6:]
    )
    try:
        rc = await simulation_controller.run()
        LOG.debug(
            "SimulationController(id=0x%x, uid=%s) exited normally.",
            id(simulation_controller),
            simulation_controller.uid,
        )
        return rc
    except Exception as e:
        LOG.critical(
            "SimulationController(id=0x%x, uid=%s) encountered a fatal "
            "error in run(): %s",
            id(simulation_controller),
            simulation_controller.uid,
            e,
        )
        tb = io.StringIO()
        traceback.print_exc(file=tb)
        LOG.debug(
            "SimulationController(id=0x%x, uid=%s) encountered a fatal "
            "error in run(): %s",
            id(simulation_controller),
            simulation_controller.uid,
            e,
            tb.getvalue(),
        )
        raise e


class RGSStartingSimControllers(RunGovernorState):
    """Represent the STARTING_SIM_CONTROLLERS state of the run
    governor.

    This state is entered after STARTING_SETUP succeeds. During
    this state, processes for all required simulation controllers are
    created (but not started).

    Possible next states are
    * :class:`.RGSStartingEnvConductors in the normal case
    * :class:`.RGSErrorHandlingStarting in the error case.

    Parameters
    ----------
    rgc: :class:`.RunGovernor`
        The run governor instance that provides the context for this
        state.

    """

    def __init__(self, rgc: "RunGovernor"):
        super().__init__(rgc, "STARTING_SIM_CONTROLLERS")

    async def run(self):
        self._init_sim_controllers()

    def next_state(self):
        from . import RGSErrorHandlingStarting, RGSStartingEnvConductors

        if len(self.rgc.errors) > 0:
            self.rgc.state = RGSErrorHandlingStarting(self.rgc)
        else:
            self.rgc.state = RGSStartingEnvConductors(self.rgc)

    def _init_sim_controllers(self):
        """Initialize all given simulation controllers as subprocesses.

        All objects of type :py:class:`SimulationController` are
        treated as subprocesses. As long as they run on the same
        machine, they are initialized in this method.

        The processes are not started.
        """
        sim_controllers = list(
            self.rgc.experiment_run.simulation_controllers(
                self.rgc.current_phase
            ).values()
        )
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) initializing %d "
            "simulation controller(s).",
            id(self.rgc),
            self.rgc.uid,
            len(sim_controllers),
        )

        def _make_process(sc):
            sc.experiment_run_id = self.rgc.experiment_run_id
            return aiomultiprocess.Process(
                name=f"SimulationController-{sc.uid}",
                target=spawn_wrapper,
                args=(
                    f"SimulationController-{sc.uid}",
                    RuntimeConfig().to_dict(),
                    _run_simulation_controller,
                    [sc],
                ),
            )

        self.rgc.sim_controllers = {
            sc.uid: _make_process(sc) for sc in sim_controllers
        }



from palaestrai.core.protocol import EnvironmentUpdateResponse
from palaestrai.experiment import TerminationCondition


class VanillaSimControllerTerminationCondition(TerminationCondition):
    """
    This is the Vanilla Simulation Controller Termination Condition.
    The Termination will return True if at least one Environment
    terminates.
    """

    def check_termination(self, message, component=None):
        if isinstance(message, EnvironmentUpdateResponse):
            if message.is_terminal:
                return True
            else:
                return False
